#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import os
import platform
import subprocess
import sys
import tempfile
from packaging.version import Version
from typing import Final, Optional
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import h5py
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def findGitRoot() -> Optional[str]:
    """ Attempt to find the git root
    """
    try:
        result = subprocess.run(['git', 'rev-parse', '--show-toplevel'],
                                capture_output=True, text=True, check=True)
        return result.stdout.strip()
    except subprocess.CalledProcessError:
        return None


def downloadGitDir(user    : str,
                   repo    : str,
                   path    : str,
                   target  : str,
                   token   : Optional[str]  = None,
                   branch  : str            = 'main',
                   progress: Optional[bool] = True) -> None:
    # Standard libraries -----------------------------------
    import json
    import time
    import urllib.request
    from urllib.error import HTTPError
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    from pyhope.common.common_progress import ProgressBar
    # ------------------------------------------------------

    # Helper to manage API requests and rate limiting
    def _make_request(url : str,
                      base: Optional[str]         = None,
                      bar : Optional[ProgressBar] = None):

        headers = {}
        if token:
            headers['Authorization'] = f'token {token}'

        req = urllib.request.Request(url, headers=headers)

        while True:
            try:
                return urllib.request.urlopen(req)
            except HTTPError as e:
                # Check for rate-limiting error
                if  e.code == 403                        \
                and 'X-RateLimit-Remaining' in e.headers \
                and int(e.headers['X-RateLimit-Remaining']) == 0:  # noqa: E271
                    timeReset = int(e.headers['X-RateLimit-Reset'])
                    timeWait  = max(timeReset - time.time(), 1)
                    if base is not None and bar is not None:
                        bar.title(f'│ Rate limited, waiting {timeWait} sec')
                    time.sleep(timeWait)
                    if base is not None and bar is not None:
                        bar.title( '│               Downloading tests')
                    # Retry the request
                    continue
                else:
                    # Re-raise other HTTP errors
                    raise

    apiURL = f'https://api.github.com/repos/{user}/{repo}/contents/{path}?ref={branch}'

    with _make_request(apiURL) as u:
        contents = json.loads(u.read().decode())

    # Exlude all tutorials with index 5 or higher
    # > These are only used for internal testing
    if progress:
        contents = [s for s in contents if s['name'][0] in '1234']

    # If we are in a subdirectory, create it
    os.makedirs(target, exist_ok=True)

    bar = None
    if progress:
        bar = ProgressBar(value=len(contents), title='│               Downloading tests', length=33, threshold=1)

    for item in contents:
        name     = item['name']
        subPath  = os.path.join(target, name)
        itemType = item.get('type')

        match itemType:
            case 'file':
                # Initially, download the file content
                # > This might be the actual file or an LFS pointer
                with _make_request(item['download_url'], os.path.basename(path), bar) as u:
                    content = u.read()

                # Check if the content is a Git LFS pointer
                if content.startswith(b'version https://git-lfs.github.com/spec/v1'):
                    # If it's an LFS pointer, the actual file needs to be downloaded
                    # from the media URL, which is constructed from the file's path
                    # print(f'Downloading LFS file: {item["path"]}...')
                    lfs_url = f'https://media.githubusercontent.com/media/{user}/{repo}/{branch}/{item["path"]}'
                    with _make_request(lfs_url, os.path.basename(path), bar) as lfs_u:
                        content = lfs_u.read()

                # Write the final content (either regular file or LFS file) to disk
                with open(subPath, 'wb') as f:
                    f.write(content)

            case 'dir':
                # Recursively call the function for subdirectories
                # > Progress is disabled for sub-calls to issues with duplicate progressBar
                downloadGitDir(user, repo, item['path'], subPath, branch=branch, progress=False)

            case _:
                print(hopout.warn(f'Unknown item type "{itemType}" for item "{name}". Skipping.'))

        if progress:
            bar.step()

    if progress:
        bar.close()


def hdf5Stats(obj: h5py.Dataset) -> Optional[dict[str, float]]:
    """ Calculate mean, min/max and standard deviation of a given HDF5 dataset
    """
    stats = {}

    if isinstance(obj, h5py.Dataset):
        try:
            data = obj[()]
            if np.issubdtype(data.dtype, np.number):
                minv    = float(np.min(data))
                maxv    = float(np.max(data))
                meanv   = float(np.mean(data))
                stddevv = float(np.std(data))
                stats = {'min': minv, 'max': maxv, 'mean': meanv, 'stddev': stddevv}
        except Exception:
            # raise ValueError('Failed to parse generated HDF5 file')
            return None

    return stats


def CheckInstall(path: Optional[str] = None) -> None:
    """ Verify the installation by comparing against known results
    """
    # Third-party libraries --------------------------------
    import h5py
    from contextlib import redirect_stdout
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    from pyhope.common.common_progress import ProgressBar
    from pyhope.readintools.readintools import ReadConfig
    # ------------------------------------------------------

    # Standard library since Python 3.11
    if Version(platform.python_version()) >= Version('3.11'):
        import tomllib  # ty: ignore[unresolved-import]
    else:
        tomllib = None

    testDir: Final[str]   = 'tutorials'
    testArr: Final[tuple] = ('ElemInfo', 'GlobalNodeIDs', 'NodeCoords', 'SideInfo')
    testLen: Final[int]   = max(len(s) for s in testArr)
    tmpDir = None

    hopout.small_banner('Verifying installation')

    # Path is None, search for the tutorials directory relative to the git root
    if not path:
        root = findGitRoot()

        if root:
            path = os.path.join(root, testDir)

    # Directory not exist, download to temporary path
    if not path or not os.path.isdir(path):
        # --- Token Discovery ---
        token = os.environ.get('GITHUB_TOKEN')
        if not token:
            try:
                # Try to get the token from the gh-cli
                token = subprocess.check_output(['gh', 'auth', 'token']).strip().decode()
                hopout.info('Discovered GitHub token, will use it for authentification')
                hopout.sep()
            except (FileNotFoundError, subprocess.CalledProcessError):
                # gh-cli not found or not logged in
                token = None
                print(hopout.warn('No GitHub token found. Using unauthenticated requests which are severely rate-limited.'))
                hopout.sep()

        tmpDir = tempfile.TemporaryDirectory(delete=False)  # pyright: ignore[reportCallIssue] # ty: ignore[no-matching-overload]
        downloadGitDir('hopr-framework', 'PyHOPE', testDir, tmpDir.name, token)
        path = tmpDir.name
        hopout.sep()

    # Final check, are there tutorials at the given path
    if not path:
        return None

    # Loop over the given path
    if os.path.isdir(path):
        tutorials = sorted(os.listdir(path))

        # Exlude all tutorials with index 5 or higher
        # > These are only used for internal testing
        tutorials = [s for s in tutorials if s[0] in '1234']

        tsuccess  = np.ones(len(tutorials), dtype=bool)

        # Enter each tutorial and look for the parameter.ini
        bar = ProgressBar(value        = len(tutorials),                       # noqa: E251
                          title        = '│                   Running tests',  # noqa: E251
                          length       = 33,                                   # noqa: E251
                          threshold    = 1,                                    # noqa: E251
                          enrich_print = False)                                # noqa: E251
        for tNum, tutorial in enumerate(tutorials):
            tutorialPath = os.path.join(path, tutorial)

            # Skip the test if the parameter file is missing
            if not os.path.isfile(os.path.join(tutorialPath, 'parameter.ini')):
                bar.step()
                continue

            # Assemble the parameter path
            parameter = os.path.join(tutorialPath, 'parameter.ini')

            # Suppress output to standard output
            try:
                with open(os.devnull, 'w') as null, redirect_stdout(null):
                    # All code that should have silent stdout here
                    with ReadConfig(parameter) as rc:
                        params = rc
            except Exception:
                # Config read failed
                bar.step()
                continue

            # Validate expected config fields
            projectName = None
            try:
                projectName = params['general']['projectname']
            except Exception:
                bar.step()
                continue

            # Run pyhope to generate mesh; trap errors and timeouts
            try:
                subprocess.run([sys.executable, '-m', 'pyhope', 'parameter.ini', '--skip-checks'],
                                cwd     = tutorialPath,        # noqa: E251
                                check   = True,                # noqa: E251
                                stdout  = subprocess.DEVNULL,  # noqa: E251
                                stderr  = subprocess.DEVNULL,  # noqa: E251
                                timeout = 300)                 # noqa: E251 # timeout in seconds
                # Alternative implementation with Popen
                # p = subprocess.Popen([sys.executable, '-m', 'pyhope', 'parameter.ini', '--skip-checks'],
                #                 cwd     = tutorialPath,        # noqa: E251
                #                 stdout  = subprocess.PIPE,     # noqa: E251
                #                 stderr  = subprocess.DEVNULL)  # noqa: E251
                # try:
                #     # if this returns, the process completed
                #     p.wait(timeout=300)
                # except subprocess.TimeoutExpired:
                #     p.terminate()
            except subprocess.CalledProcessError as exc:
                tsuccess[tNum] = False
                hopout.info(f'{hopout.Symbols.ERR } PyHOPE failed for "{tutorial}": ' +
                               f'Return code = {getattr(exc, "returncode", "")}')
                bar.step()
                continue
            except subprocess.TimeoutExpired:
                tsuccess[tNum] = False
                hopout.info(f'{hopout.Symbols.ERR } PyHOPE timed out for "{tutorial}"')
                bar.step()
                continue
            except Exception as exc:
                tsuccess[tNum] = False
                hopout.info(f'{hopout.Symbols.ERR } Unexpected error running PyHOPE for "{tutorial}": {exc}')
                bar.step()
                continue

            # Check if PyHOPE generated an output file
            if not os.path.isfile(os.path.join(tutorialPath, f'{projectName}_mesh.h5')):
                # raise ExecError(f'PyHOPE failed to generate mesh for {tutorial}')
                tsuccess[tNum] = False
                hopout.info(f'{hopout.Symbols.ERR } PyHOPE did not produce {projectName}_mesh.h5 for "{tutorial}"')
                bar.step()
                continue

            # Open the TOML file
            tomlData  = None
            toml_path = os.path.join(tutorialPath, 'analyze.toml')
            if os.path.isfile(toml_path):
                try:
                    with open(toml_path, mode='rb') as f:
                        tomlData = tomllib.load(f)
                except Exception:
                    # If TOML is present but invalid,
                    # Python 3.11+: Skip the tutorial
                    # Python 3.10-: Mark the tutorial as successful
                    bar.step()
                    if Version(platform.python_version()) < Version('3.11'):
                        hopout.info(f'{hopout.Symbols.OK  } Successfully completed test (w/o verification) "{tutorial}"')
                        tsuccess[tNum] = True
                    continue

            # Open the mesh file and compare against the reference
            with h5py.File(os.path.join(tutorialPath, f'{projectName}_mesh.h5'), mode='r') as h5:
                for key, val in h5.items():
                    # Calculate the padding length
                    padLen = max(0, testLen - len(key))

                    # Only relevant datasets
                    if key not in testArr:
                        continue

                    # Load the stats from the HDF5 file
                    try:
                        h5stats = hdf5Stats(val)
                    except Exception as exc:
                        tsuccess[tNum] = False
                        hopout.routine(f'{hopout.Symbols.ERR } Failed computing stats for {key} in "{tutorial}": {exc}')
                        # Continue to next dataset
                        continue

                    # Load the stats from the TOML file
                    if  h5stats  is not None \
                    and tomlData is not None \
                    and key in tomlData.keys():  # noqa: E271, E272
                        # Fallback tolerances:
                        if 'GlobalNodeIDs' in key:
                            # GlobalNodeIDs are susceptible to rounding issues
                            rtol = tomlData.get('_defaults', {}).get('rtol', 1E-6)
                            atol = tomlData.get('_defaults', {}).get('atol', 1E-1)
                        else:
                            rtol = tomlData.get('_defaults', {}).get('rtol', 1E-6)
                            atol = tomlData.get('_defaults', {}).get('atol', 1E-8)
                        # If dataset-specific tolerances exist, use them
                        ds_defaults = tomlData.get(key, {})
                        rtol = ds_defaults.get('_rtol', rtol)
                        atol = ds_defaults.get('_atol', atol)

                        for skey, sval in tomlData[key].items():
                            if skey.startswith('_'):
                                # skip metadata keys starting with underscore
                                continue

                            # Ensure the stat exists in h5stats
                            if skey not in h5stats:
                                tsuccess[tNum] = False
                                hopout.routine(f'{hopout.Symbols.ERR } Missing stat "{skey}" for {key} in "{tutorial}"')
                                continue

                            hval = h5stats[skey]
                            # Finally, compare the values
                            try:
                                # If both are numeric scalars
                                if np.isscalar(hval) and np.isscalar(sval):
                                    if not np.isclose(hval, sval, rtol=rtol, atol=atol, equal_nan=True):
                                        tsuccess[tNum] = False
                                        hopout.routine(f'{hopout.Symbols.ERR } Stat mismatch for "{tutorial}" {key}.{skey}: ' +
                                                    f'h5={hval} toml={sval}')
                                else:
                                    # Try comparing as arrays
                                    h_arr = np.asarray(hval)
                                    s_arr = np.asarray(sval)

                                    # Shapes should match
                                    if h_arr.shape != s_arr.shape:
                                        tsuccess[tNum] = False
                                        hopout.routine(f'{hopout.Symbols.ERR } Shape mismatch for "{tutorial}" {key}.{skey}: ' +
                                                    f'h5={h_arr.shape} toml={s_arr.shape}')
                                        continue

                                    # Values should match
                                    if not np.allclose(h_arr, s_arr, rtol=rtol, atol=atol, equal_nan=True):
                                        tsuccess[tNum] = False
                                        hopout.routine(f'{hopout.Symbols.ERR } Array stat mismatch for "{tutorial}" {key}.{skey}')

                            except Exception as exc:
                                tsuccess[tNum] = False
                                hopout.routine(f'{hopout.Symbols.ERR } Error comparing stat "{skey}" for "{tutorial}": {exc}')
                                continue

                    else:
                        # No TOML entry for this dataset
                        tsuccess[tNum] = False
                        hopout.routine(f'{hopout.Symbols.ERR } No TOML stats for dataset "{key}"{" " * padLen} in "{tutorial}"')

            bar.step()

            if tsuccess[tNum]:
                hopout.info(f'{hopout.Symbols.OK  } Successfully completed test "{tutorial}"')
            else:
                hopout.info(f'{hopout.Symbols.ERR } Failed test "{tutorial}"')

        # Cleanup temporary directory
        if  tmpDir is not None             \
        and os.path.isdir(tmpDir.name)     \
        and tmpDir.name.startswith('/tmp'):  # noqa: E271
            tmpDir.cleanup()

        # Close the progress bar
        if bar is not None:
            hopout.sep()
            bar.title('│          Finished running tests')
            bar.close()
        hopout.info('')

        # Print the final output
        hopout.small_banner('Verification summary')
        if all(tsuccess):
            hopout.info(f'{hopout.Symbols.OK  } Successfully completed {np.sum(tsuccess)}/{len(tutorials)} tests')
        else:
            hopout.info(f'{hopout.Symbols.ERR } Failed {np.sum(~tsuccess)}/{len(tutorials)} tests')
        hopout.separator(length=79)

        # Final exit code
        if not all(tsuccess):
            sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import re
import sys
from typing import Final, Optional, cast
from collections.abc import Iterable
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Typing libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import typing
if typing.TYPE_CHECKING:
    import numpy.typing as npt
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
import pyhope.mesh.mesh_vars as mesh_vars
from pyhope.mesh.mesh_common import sidetovol2
from pyhope.mesh.mesh_common import face_to_nodes
# ==================================================================================================================================


def check_sides(elem,
                failed_only: bool = False,
               ) -> Optional[list[tuple]]:
    """ Check if connected sides have matching corner nodes
    """
    results = None
    elems:  Final[list]  = mesh_vars.elems
    sides:  Final[list]  = mesh_vars.sides
    nGeo:   Final[int]   = mesh_vars.nGeo
    bcs:    Final[list]  = mesh_vars.bcs
    # Tolerance for physical comparison
    tol:    Final[float] = mesh_vars.tolPeriodic
    vvs:    Final[list]  = mesh_vars.vvs
    points: Final[npt.NDArray] = mesh_vars.mesh.points

    for SideID in elem.sides:
        master = sides[SideID]
        # Only connected sides that are master sides and not small mortar sides
        if master.connection is None \
        or master.connection < 0     \
        or master.MS != 1            \
        or master.sideType < 0:  # noqa: E271
            continue

        side   = (master, sides[master.connection])

        # Only actual element sides
        if side[0].face is None      \
        or side[1].face is None:  # noqa: E271
            continue

        # Sanity check the flip with the other nodes
        elem0  = elems[side[0].elemID]
        elem1  = elems[side[1].elemID]
        if elem1.type % 100 != 8:
            continue

        # Map the meshio nodes to the tensor-product nodes
        elemType = elem0.type
        nodes    = elem0.nodes[sidetovol2(nGeo, 0           , side[0].face, elemType)]
        nbNodes  = elem1.nodes[sidetovol2(nGeo, side[1].flip, side[1].face, elemType)]

        # INFO: THIS CURRENTLY MIGHT NOT WORK SINCE WE POTENTIALLY ONLY HAVE THE CORNER NODES AVAILABLE
        try:
            # Translate to periodic nodes if required
            if side[0].bcid is not None and side[1].bcid is not None and bcs[side[1].bcid].type[0] == 1:
                nbNodes = np.vectorize(lambda s: mesh_vars.periNodes[(s, bcs[side[1].bcid].name)], otypes=[int])(nbNodes)
            # Check if the node IDs match
            success = np.array_equal(nodes, nbNodes)
        # Fallback to comparison of physical coordinates
        except KeyError:
            # Check if periodic vector matches using vectorized np.allclose
            iVV = bcs[side[0].bcid].type[3]
            vv  = vvs[np.abs(iVV) - 1]['Dir'] * np.sign(iVV)
            success = np.allclose(points[nodes] + vv, points[nbNodes], rtol=tol, atol=tol)

        # If requested, only return errors
        if failed_only and success:
            continue

        # Lazily initialize results on first failure
        if results is None:
            results = []
        results.append((success, SideID))

    # Avoid creating empty lists on elem_results
    if results is None:
        return None if failed_only else []

    return results


def process_chunk(chunk) -> list:
    """Process a chunk of elements by checking surface normal orientation
    """
    # Only keep failures to reduce memory and avoid building large arrays of successes
    chunk_results = []
    for elem in chunk:
        elem_results = check_sides(elem,
                                   failed_only=True)
        # Append a lightweight sentinel (None) for successes, actual failure list otherwise
        chunk_results.append(elem_results)
    return chunk_results


def CheckConnect() -> None:
    """ Check if the mesh is correctly connected
    """
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    import pyhope.mesh.mesh_vars as mesh_vars
    from pyhope.common.common_parallel import run_in_parallel
    from pyhope.common.common_vars import np_mtp
    from pyhope.readintools.readintools import GetLogical
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('CHECK CONNECTIVITY...')
    hopout.sep()

    checkConnectivity  = GetLogical('CheckConnectivity')
    if not checkConnectivity:
        return None

    # Check all sides
    elems:     Final[list] = mesh_vars.elems

    # Only consider hexahedrons
    if any(cast(int, e.type) % 100 != 8 for e in elems):
        elemTypes = list(set([e.type for e in elems if e.type % 100 != 8]))
        print(hopout.warn('Ignored element type: {}'.format(
            [re.sub(r"\d+$", "", mesh_vars.ELEMTYPE.inam[e][0]) for e in elemTypes]
        )))
        return

    # Prepare elements for parallel processing
    if np_mtp > 0:
        # Run in parallel with a chunk size
        # > Dispatch the tasks to the workers, minimum 10 tasks per worker, maximum 1000 tasks per worker
        res     = run_in_parallel(process_chunk,
                                  elems,
                                  chunk_size=max(1, min(1000, max(10, int(len(elems)/(40.*np_mtp))))),
                                 )
    else:
        res     = [elem for elem in elems if check_sides(elem, failed_only=True)]

    if len(res) > 0:
        # Flatten per-element results (skip None placeholders)
        results = tuple(result for elem_results in res if isinstance(elem_results, Iterable) and elem_results is not None
                               for result       in elem_results)  # noqa: E272

        nGeo:      Final[int]        = mesh_vars.nGeo
        sides:     Final[list]       = mesh_vars.sides
        points:    Final[np.ndarray] = mesh_vars.mesh.points

        # Compute total number of checked connections without materializing all results
        nconn = 0
        for SideID, side in enumerate(sides):
            # Only connected sides and not small mortar sides
            if side.connection is None or side.sideType < 0:
                continue
            # Big mortar side is counted once
            elif side.connection < 0:
                nconn += 1
            # Internal side: only count the canonical representative and ignore virtual mortar sides
            elif side.connection >= 0:
                if SideID > side.connection:
                    continue
                if side.locMortar is not None:
                    continue
                nconn += 1

        for result in cast(tuple[tuple], results):
            # Unpack the results
            side    = sides[result[1]]
            elem    = elems[side.elemID]
            nbside  = sides[side.connection]
            nbelem  = elems[nbside.elemID]

            nodes   =   elem.nodes[face_to_nodes(  side.face,   elem.type, nGeo)]
            nbnodes = nbelem.nodes[face_to_nodes(nbside.face, nbelem.type, nGeo)]

            print()
            # Check if side is oriented inwards
            errStr = 'Side connectivity does not match the calculated neighbour side'
            print(hopout.warn(errStr, length=len(errStr)+16))

            # Print the information
            strLen  = max(len(str(side.sideID+1)), len(str(nbside.sideID+1)))
            print(hopout.warn(f'> Element {  elem.elemID+1:>{strLen}}, Side {  side.face}, Side {  side.sideID+1:>{strLen}}'))  # noqa: E501
            print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in points[  nodes[ 0,  0]]) + ']'))    # noqa: E271
            print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in points[  nodes[ 0, -1]]) + ']'))    # noqa: E271
            print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in points[  nodes[-1,  0]]) + ']'))    # noqa: E271
            print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in points[  nodes[-1, -1]]) + ']'))    # noqa: E271
            # print()
            print(hopout.warn(f'> Element {nbelem.elemID+1:>{strLen}}, Side {nbside.face}, Side {nbside.sideID+1:>{strLen}}'))  # noqa: E501
            print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in points[nbnodes[ 0,  0]]) + ']'))    # noqa: E271
            print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in points[nbnodes[ 0, -1]]) + ']'))    # noqa: E271
            print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in points[nbnodes[-1,  0]]) + ']'))    # noqa: E271
            print(hopout.warn('- Coordinates  : [' + ' '.join('{:12.3f}'.format(s) for s in points[nbnodes[-1, -1]]) + ']'))    # noqa: E271

        hopout.warning(f'Connectivity check failed for {len(results)} / {nconn} connections!')
        sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from collections import defaultdict
from typing import Dict, Tuple, cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


# def copysign_int(x: int, y: int) -> int:
#     """ Return a int with the magnitude (absolute value) of x but the sign of y
#     """
#     # Standard libraries -----------------------------------
#     import math
#     # ------------------------------------------------------
#     return int(math.copysign(x, y))


def FEMConnect() -> None:
    """ Generate connectivity information for edges and vertices
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.readintools.readintools import CountOption, GetLogical
    from pyhope.mesh.mesh_common import edges, edge_to_corner
    # ------------------------------------------------------

    if CountOption('doFEMConnect') == 0:
        return None

    hopout.separator()
    hopout.info('GENERATE FINITE ELEMENT METHOD (FEM) CONNECTIVITY...')
    hopout.sep()

    doFEMConnect = GetLogical('doFEMConnect')
    if not doFEMConnect:
        hopout.separator()
        return None

    elems     = mesh_vars.elems
    periNodes = mesh_vars.periNodes

    # Build a graph of all periodic connections across all BCs
    periGraph: dict[int, set[int]] = defaultdict(set)
    for (node, _bc), peri in periNodes.items():
        node, peri = int(node), int(peri)
        periGraph[node].add(peri)
        periGraph[peri].add(node)

    # Find connected components; representative = min(node indices in component)
    nodeFirst: dict[int, int]      = {}  # node -> representative (canonical)
    nodeGroup: dict[int, set[int]] = {}  # representative -> full member set

    for start in list(periGraph.keys()):
        if start in nodeFirst:
            continue

        stack = [start]
        comp  = set()

        while stack:
            cur = stack.pop()
            if cur in comp:
                continue

            comp.add(cur)
            for nxt in periGraph[cur]:
                if nxt not in comp:
                    stack.append(nxt)

        rep = min(comp)
        for v in comp:
            nodeFirst[v] = rep

        nodeGroup[rep] = comp

    # Convenience mapping: each node -> full set of periodic equivalents (including itself)
    periGroups: dict[int, set[int]] = {}
    for rep, members in nodeGroup.items():
        for v in members:
            periGroups[v] = members

    # Build mapping of each node -> set of element indices that include that node
    nodeToElements = defaultdict(set)
    for idx, elem in enumerate(elems):
        for node in cast(np.ndarray, elem.nodes)[:cast(int, elem.type) % 10]:
            nodeToElements[int(node)].add(idx)

    # Precompute combined connectivity for each node
    # > For a given node, the combined set is:
    # > nodeToElements[node] ∪ nodeToElements[periDict[node]]
    nodeConn = {node: set().union(*(nodeToElements.get(eq, set()) for eq   in periGroups.get(node, {node})))  # noqa: E272
                                                                  for node in nodeToElements.keys()}

    # Collect all unique canonical vertices from every element
    # > The canonical vertex is the minimum of the node and its periodic counterparts
    canonicalSet = {nodeFirst.get(int(node), int(node)) for elem in elems
                                                        for node in cast(np.ndarray, elem.nodes)[:cast(int, elem.type) % 10]}

    # Create a mapping from each canonical vertex to a unique index
    # > FEMVertexID starts at 1
    sortedCanonical = sorted(canonicalSet)
    FEMNodeMapping  = { canonical: newID for newID, canonical in enumerate(sortedCanonical, start=1)}

    # EDGE1: Build Per-BC DIRECTED Mappings
    # > To create a unique signature for each edge, we need to know how its nodes
    # > are displaced by each periodic boundary condition. We build a simple, directed
    # > map for each BC that directly reflects the (source -> target) relationship in
    # > periNodes. We only want to map from negative to positive, thus keep the direction
    periNames = sorted(list({bc for _, bc in periNodes.keys()}))

    # This dictionary holds the directed mapping for each BC
    # > Key: Source node
    # > Val: Target node
    nodeMapBC: dict[str, dict[int, int]] = {}
    for bc in periNames:
        nodeMapBC[bc] = {int(node): int(peri) for (node, bc_name), peri in periNodes.items() if bc_name == bc}

    # For convenience, also build a set of all nodes that lie on any boundary
    BCNodes = {int(node)      for node, _ in periNodes.keys()}     # noqa: E272
    BCNodes.update({int(peri) for peri    in periNodes.values()})  # noqa: E272

    # EDGE2: Enumerate All Raw Edges from the Mesh
    # > tuple(element_index, local_edge_index, (node0, node1))
    edgesRaw: list[tuple[int, int, tuple[int, int]]] = []

    for elemID, elem in enumerate(elems):
        for edge in edges(elem.type):
            # Get the local corner indices for the current edge
            edgeCorners = edge_to_corner(edge, elem.type)
            # Get the global node indices for those corners
            n0, n1 = int(cast(np.ndarray, elem.nodes)[edgeCorners[0]]), int(cast(np.ndarray, elem.nodes)[edgeCorners[1]])
            edgesRaw.append((elemID, edge, (n0, n1)))

    # EDGE3: Generate Canonical Edge Keys (Graph-Based Approach)
    # > This identifies all periodically equivalent edges by building a graph of their representations
    # > 1. Build an Equivalence Graph
    #      The keys are edge-representations (tuples), and the values are sets of other edge-representations they are connected to
    edgeGraph = defaultdict(set)

    for _, _, nodes in edgesRaw:
        # Start with the edge's own representation
        edgeBase = tuple(sorted(nodes))
        # Find all possible displaced representations
        edgePeri = {edgeBase}
        nodesSet = frozenset(nodes)

        # Check if all nodes are on a boundary
        if nodesSet.issubset(BCNodes):
            for bc in periNames:
                nodeMap = nodeMapBC[bc]
                if nodesSet.issubset(nodeMap) and nodeMap[nodes[0]] != nodeMap[nodes[1]]:
                    edgePeri.add(tuple(sorted([nodeMap[node] for node in nodes])))

        # Connect all representations for this edge in the graph
        # > This ensures that if A->B and B->C, we can later find that A is related to C
        edgeList = list(edgePeri)
        for i in range(len(edgeList)):
            for j in range(i + 1, len(edgeList)):
                u, v = edgeList[i], edgeList[j]
                edgeGraph[u].add(v)
                edgeGraph[v].add(u)

    # > 2. Find connected components and their canonical representatives
    #      This dict stores the single "canonical edge" for each group
    edgeCanonical = {}
    visited       = set()

    # Iterate through all unique representations
    edgeSet = set(edgeGraph.keys())
    for _, _, nodes in edgesRaw:
        edgeSet.add(tuple(sorted(nodes)))

    for nodeStart in edgeSet:
        # Done with this node
        if nodeStart in visited:
            continue

        # This component holds all edge representations for one unique FEM edge
        component = set()
        stack     = [nodeStart]

        while stack:
            currentNode = stack.pop()
            if currentNode in visited:
                continue

            visited  .add(currentNode)
            component.add(currentNode)

            for neighbor in edgeGraph.get(currentNode, []):
                if neighbor not in visited:
                    stack.append(neighbor)

        # > 3. The true canonical key is the minimum representation in the component
        canonical_rep = min(component)

        # > 4. Map all members of the component to this single canonical rep
        for node in component:
            edgeCanonical[node] = canonical_rep

    # > 5. Generate final edge keys for all raw edges
    edgeKeys = []
    for elemID, locEdge, nodes in edgesRaw:
        # Get the initial representation of the edge
        edgeBase = tuple(sorted(nodes))

        # Find the true canonical representation from the pre-computed map
        canonical_edge_nodes = edgeCanonical.get(edgeBase, edgeBase)

        c0 = nodeFirst.get(nodes[0], nodes[0])
        c1 = nodeFirst.get(nodes[1], nodes[1])
        v0 = FEMNodeMapping[c0]
        v1 = FEMNodeMapping[c1]
        edgePair = tuple(sorted((v0, v1)))

        # The final, unique key for this edge
        edgeKey = (edgePair, canonical_edge_nodes)

        # Create the edge key list
        edgeKeys.append((elemID, locEdge, edgeKey, (v0, v1), nodes))

    # Create the final mapping from the unique edge key to a simple integer ID
    uniqueEdges    = sorted(list({k for _, _, k, _, _ in edgeKeys}))
    FEMEdgeMapping = {key: i for i, key in enumerate(uniqueEdges)}

    # Build the vertex connectivity
    for idx, elem in enumerate(elems):
        vertexInfo: Dict[int, Tuple[int, Tuple[int, ...]]] = {}
        for locNode, node in enumerate(int(n) for n in cast(np.ndarray, elem.nodes)[:cast(int, elem.type) % 10]):
            # Determine canonical vertex id
            canonical   = nodeFirst.get(node, node)
            FEMVertexID = FEMNodeMapping[canonical]
            # Retrive connectivity set for the node
            nodeVertex = nodeConn.get(node, set())
            vertexInfo[locNode] = (FEMVertexID, tuple(sorted(nodeVertex)))
        # Set the vertex connectivity for the element
        elem.vertexInfo = vertexInfo

        # Initialize edgeInfo dictionaries for all elements first
        elem.edgeInfo = {}

    # Use the pre-computed edge_key_list
    for elemID, locEdge, edgeKey, edgePair, edgeNodes in edgeKeys:
        # Get the global FEMEdgeID
        FEMEdgeID = FEMEdgeMapping[edgeKey]
        # Retrieve the edge connectivity for the element
        # The structure is: (local_idx, global_id, global_vertex_ids, local_node_ids)
        elems[elemID].edgeInfo[locEdge] = (locEdge, FEMEdgeID, edgePair, edgeNodes)


def getFEMInfo(nodeInfo: np.ndarray) -> tuple[np.ndarray,  # FEMElemInfo
                                              int,         # nVertices
                                              np.ndarray,  # VertexInfo
                                              np.ndarray,  # VertexConnectInfo
                                              int,         # nEdges
                                              np.ndarray,  # EdgeInfo
                                              np.ndarray   # EdgeConnectInfo
                                             ]:
    """ Extract the FEM connectivity information and return five arrays

     - FEMElemInfo      : [offsetIndEdge, lastIndEdge, offsetIndVertex, lastIndVertex]
     - VertexInfo       : [FEMVertexID, offsetIndVertexConnect, lastIndVertexConnect]
     - VertexConnectInfo: [nbElemId, nbLocVertexId]
     - EdgeInfo         : [FEMEdgeID,   offsetIndEdgeConnect,   lastIndEdgeConnect]
     - EdgeConnectInfo  : [nbElemID, nbLocEdgeID]
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------

    elems  = mesh_vars.elems
    nElems = len(elems)

    # Check if elements contain FEM connectivity
    if not hasattr(elems[0], 'vertexInfo') or elems[0].vertexInfo is None:
        return np.array([]), 0, np.array([]), np.array([]), 0, np.array([]), np.array([])

    # Vertex connectivity info ---------------------------------------------------
    # > Build list of all vertex occurrences, appearing in the same order as the elements
    occList = [(FEMVertexID, elemID, locNode) for elemID , elem             in enumerate(elems)  # noqa: E272
                                              for locNode, (FEMVertexID, _) in elem.vertexInfo.items()]
    nFEMVertices = max(FEMVertexID for FEMVertexID, _, _ in occList)

    # > Build mapping from FEM vertex ID to list of occurrences
    groups = defaultdict(list)
    for occIdx, (FEMVertexID, elemID, locNode) in enumerate(occList):
        groups[FEMVertexID].append((occIdx, elemID, locNode))

    # Initialize FEM element information
    FEMElemInfo    = np.zeros((nElems, 4), dtype=np.int32)

    vertexInfoList = []  # List: [FEMVertexID, offsetIndVertexConnect, lastIndVertexConnect]
    vertexConnList = []  # List: [[nbElemID, nbLocVertexID]]
    vertexOffset   = 0
    occGlobalIdx   = 0   # global index in occList

    for elemID, elem in enumerate(elems):
        # Process vertex occurrences for the current element
        for _ in cast(dict, elem.vertexInfo):
            # Get the occurrence information from the global occList
            FEMVertexID, _, locNode = occList[occGlobalIdx]
            groupOcc = groups[FEMVertexID]
            offset   = len(vertexConnList)

            # Identify the master occurrence (lowest occIdx from the occurrence group)
            masterOcc = min(x[0] for x in groupOcc)

            # Build connectivity list for current element, excluding itself
            connections = [(nbElem+1, nbLocal+1) if   otherOcc == masterOcc else (-(nbElem+1), nbLocal+1)  # noqa: E271
                                                 for (otherOcc, nbElem, nbLocal) in groupOcc if otherOcc != occGlobalIdx]

            if connections:
                lastIndex = offset + len(connections)
                vertexConnList.extend(connections)
            else:  # No connections
                lastIndex = offset

            # Append vertex information
            vertexInfoList.append([FEMVertexID, offset, lastIndex])
            occGlobalIdx += 1

        # Set the vertex connectivity offset for this element.
        FEMElemInfo[elemID, 2] = vertexOffset
        FEMElemInfo[elemID, 3] = vertexOffset + len(cast(dict, elem.vertexInfo))
        vertexOffset += len(cast(dict, elem.vertexInfo))

    # Edge   connectivity info ---------------------------------------------------
    # > Build list of all raw edge occurrences, appearing in the same order as the elements
    occList = [{'EdgeID': edgeIdx, 'elem': elemID, 'loc': locEdge, 'nodes': edgeNodes} for elemID, elem                        in enumerate(elems)        # noqa: E272, E501
                                                                                       for locEdge, (_, edgeIdx, _, edgeNodes) in elem.edgeInfo.items()]  # noqa: E501
    # > EdgeID starts at zero, so add 1
    nFEMEdges = max(d['EdgeID'] for d in occList) + 1

    # 2. Group these occurrences by their canonical FEMEdgeID
    groups = defaultdict(list)
    for occ in occList:
        groups[occ['EdgeID']].append(occ)

    edgeInfoList   = []  # List: [FEMEdgeID, offsetIndEdgeConnect, lastIndEdgeConnect]
    edgeConnList   = []  # List: [[nbElemID, nbLocEdgeID]]
    edgeOffset     = 0
    occGlobalIdx   = 0   # global index in occList

    for elemID, elem in enumerate(elems):
        # Process edge occurrences for the current element
        for _ in range(len(cast(dict, elem.edgeInfo))):
            # Get the occurrence information from the global occList
            currentEdge = occList[occGlobalIdx]
            FEMEdgeID   = currentEdge['EdgeID']

            # Get all siblings (including periodic ones) from the edge group
            groupOcc = groups[FEMEdgeID]
            offset   = len(edgeConnList)

            # Identify the master occurrence (lowest occIdx from the occurrence group)
            # WARNING: The master/slave logic is simplified here. We assume the first occurrence in the list is the master
            masterOcc = groupOcc[0]

            # Build connectivity list for current element, excluding itself
            connections = []
            for sibling in groupOcc:
                if sibling['elem'] == currentEdge['elem'] and sibling['loc'] == currentEdge['loc']:
                    continue

                edgeIsMaster  = (sibling['elem'] == masterOcc['elem'] and sibling['loc'] == masterOcc['loc'])
                # TODO: Check if the orientation of the master edge is with ascending nodeInfo index
                orientation   = 1 if nodeInfo[masterOcc['nodes'][0]] < nodeInfo[masterOcc['nodes'][1]] else -1

                # The current edge is the master
                # if masterID == -1:
                #     orientedElemID  = -(nbElem   +1)
                #     orientedLocEdge =   nbLocEdge+1 if nbEdge   == masterEdge               else -(nbLocEdge+1)  # noqa: E272
                if edgeIsMaster:
                    orientedElemID  = -(sibling['elem'] + 1)
                    orientedLocEdge =   sibling['loc']  + 1

                # Current edge is a slave edge
                # # The master edge is one of the connections, indicated by masterID
                else:
                    # orientedElemID  =   nbElem   +1 if masterID == iConn                    else -(nbElem   +1)  # noqa: E272
                    # orientedLocEdge =   nbLocEdge+1 if nbEdge   == connections[masterID][3] else -(nbLocEdge+1)
                    # Check our relative orientation
                    orientation     = orientation if nodeInfo[sibling['nodes'][0]] == nodeInfo[masterOcc['nodes'][0]] else -1
                    orientedElemID  =  sibling['elem'] + 1
                    orientedLocEdge =  sibling['loc']  + 1

                # TODO: Check if this is correct
                # > Copy the orientation
                orientedLocEdge = int(orientedLocEdge * orientation)

                connections.append([orientedElemID, orientedLocEdge])

            if connections:
                lastIndex = offset + len(connections)
                edgeConnList.extend(connections)
            else:
                lastIndex = offset

            # Append edge information
            edgeInfoList.append([FEMEdgeID, offset, lastIndex])
            occGlobalIdx += 1

        # Set the edge connectivity offset for this element
        FEMElemInfo[elemID, 0] = edgeOffset
        FEMElemInfo[elemID, 1] = edgeOffset + len(cast(dict, elem.edgeInfo))
        edgeOffset += len(cast(dict, elem.edgeInfo))

    # Convert lists to numpy arrays
    vertexInfo = np.array(vertexInfoList, dtype=np.int32)
    vertexConn = np.array(vertexConnList, dtype=np.int32) if vertexConnList else np.array((0, 2), dtype=np.int32)

    edgeInfo   = np.array(edgeInfoList  , dtype=np.int32)
    edgeConn   = np.array(edgeConnList  , dtype=np.int32) if edgeConnList   else np.array((0, 2), dtype=np.int32)  # noqa: E272

    return FEMElemInfo, nFEMVertices, vertexInfo, vertexConn, nFEMEdges, edgeInfo, edgeConn



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import gc
from collections import defaultdict
from functools import cache
from typing import Any, Union, cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import meshio
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
import pyhope.mesh.mesh_vars as mesh_vars
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# Instantiate ELEMTYPE
elemTypeClass = mesh_vars.ELEMTYPE()
# ==================================================================================================================================


@cache
def NDOFperElemType(elemType: str, nGeo: int) -> int:
    """ Calculate the number of degrees of freedom for a given element type
    """
    match elemType:
        case _ if elemType.startswith('triangle'):
            return round((nGeo+1)*(nGeo+2)/2.)
        case _ if elemType.startswith('quad'):
            return round((nGeo+1)**2)
        case _ if elemType.startswith('tetra'):
            return round((nGeo+1)*(nGeo+2)*(nGeo+3)/6.)
        case _ if elemType.startswith('pyramid'):
            return round((nGeo+1)*(nGeo+2)*(2*nGeo+3)/6.)
        case _ if elemType.startswith('wedge'):
            return round((nGeo+1)**2 *(nGeo+2)/2.)
        case _ if elemType.startswith('hexahedron'):
            return round((nGeo+1)**3)
        case _:
            raise ValueError(f'Unknown element type {elemType}')


@cache
def gambit_faces(elemType: Union[int, str]) -> list[str]:
    """ Return a list of all sides of an element
    """
    faces_map = {  # Tetrahedron
                   # Pyramid
                   # Wedge / Prism
                   # Hexahedron
                   8: ['y-', 'x+', 'y+', 'x-', 'z-', 'z+']
                }

    if isinstance(elemType, str):
        elemType = elemTypeClass.name[elemType]

    if elemType % 100 not in faces_map:
        raise ValueError(f'Error in faces: elemType {elemType} is not supported')

    return faces_map[elemType % 100]


def ReadGambit(fnames: list, mesh: meshio.Mesh) -> meshio.Mesh:
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    # import pyhope.mesh.mesh_vars as mesh_vars  # Already imported at the top
    from pyhope.common.common import lines_that_contain
    from pyhope.mesh.mesh_common import face_to_cgns
    from pyhope.mesh.mesh_common import FaceOrdering
    from pyhope.meshio.meshio_ordering import NodeOrdering
    # ------------------------------------------------------

    hopout.sep()

    # Create an empty meshio object
    points   = mesh.points if len(mesh.points.shape)>1 else np.zeros((0, 3), dtype=np.float64)
    pointl   = cast(list, points.tolist())
    cells    = mesh.cells_dict
    cellsets = defaultdict(lambda: defaultdict(list))

    nodeCoords   = mesh.points
    nSides       = np.zeros(2, dtype=int)
    elemTypes    = []

    # Initialize the node ordering
    node_ordering = NodeOrdering()

    for fnum, fname in enumerate(fnames):
        # Check if the file is using ASCII format internally
        with open(fname, 'r') as f:
            # Check if the file is in ASCII format
            try:
                # Read the file content
                content   = f.readlines()
                useBinary = not any('CONTROL INFO' in line for line in content)
            except UnicodeDecodeError:
                raise ValueError('Gambit binary files are not implemented yet')

            if not useBinary:
                # Search for the line containing the number of elements
                elemLine = lines_that_contain('NUMNP'            , content)[0] + 1
                # Read and unpack the number of elements
                npoints, nelems, _, nbcs, nDim, _ = map(int, content[elemLine].strip().split())

                # PyHOPE currently only supports 3D meshes
                if nDim != 3:
                    raise ValueError(f'Unsupported mesh dimension {nDim}. Only 3D meshes are supported.')

                # Check if the number of boundary conditions match the parameter file
                if nbcs > len(mesh_vars.bcs):
                    hopout.error(f'Number of boundary conditions in the mesh ({nbcs}) ' +
                                 f'does not match the parameter file ({len(mesh_vars.bcs)})')

                # Search for the line starting the node coordinates
                nodeLine = lines_that_contain('NODAL COORDINATES', content)[0]

                # Iterate and unpack the node coordinates
                nodeCoords = content[nodeLine+1:nodeLine+npoints+1]
                nodeCoords = np.genfromtxt(nodeCoords, dtype=np.float64, delimiter=None, usecols=(1, 2, 3))
                pointl.extend(nodeCoords)

                # Search for the line starting the element connectivity
                elemLine = lines_that_contain('ELEMENTS/CELLS', content)[0] + 1
                elemIter = iter(content[elemLine:])

                # Iterate and unpack the element connectivity
                for line in elemIter:
                    if 'ENDOFSECTION' in line:
                        break

                    tokens = line.strip().split()
                    if not tokens:
                        continue

                    try:
                        elemID, gType, nNodes, *elemNodes = tokens
                        elemID, gType, nNodes             = int(elemID), int(gType), int(nNodes)
                    except ValueError:
                        continue

                    # Map gambit element type to meshio element type
                    elemType  = node_ordering.typing_gambit_to_meshio(gType)
                    elemTypes.append(elemType)

                    # Check if the number of nodes matched the expected number
                    if nNodes != NDOFperElemType(elemType, mesh_vars.nGeo):
                        hopout.error(f'Number of element nodes ({nNodes}) does not match expectation for NGeo={mesh_vars.nGeo}')

                    # Keep extending the element connectivity until elemNodes is reached
                    while len(elemNodes) < nNodes:
                        elemNodes.extend(next(elemIter).strip().split())

                    # Convert elemNodes to a numpy array of integers
                    elemNodes = np.array(elemNodes, dtype=np.uint64)
                    elemNodes = node_ordering.ordering_gambit_to_meshio(elemType, elemNodes) - 1

                    cells.setdefault(elemType, []).append(elemNodes.astype(np.uint64))

                # Clean-up for memory safety
                del elemLine
                del elemIter

                # Check if the number of elements match the header
                if nelems != sum(len(cells[key]) for key in cells):
                    raise ValueError('Failed to obtain the correct number of elements.')

                # Search for the line starting the element groups
                grsLine  = lines_that_contain('ELEMENT GROUP', content)[0] + 1
                grsIter  = iter(content[grsLine:])
                lnum     = 0

                # Iterate and unpack the boundary conditions
                for line in grsIter:
                    lnum += 1
                    # Iterate until the number of boundary conditions is reached
                    if 'ENDOFSECTION' in line:
                        # Check if the next sections is also an element group
                        if 'ELEMENT GROUP' not in content[grsLine+lnum]:
                            break

                    tokens = line.strip().split()
                    if not tokens:
                        continue

                    try:
                        _, zoneNum, _, zoneNElems, _, zoneMat, _, zoneFlags = tokens
                        zoneNum, zoneNElems, zoneMat, zoneFlags = int(zoneNum), int(zoneNElems), int(zoneMat), int(zoneFlags)
                    except ValueError:
                        continue

                    # When merging grids with zoneID = 1, we want them to have separate IDs after the merge
                    # FIXME: GAMBIT allows own volume zone, we should read them here
                    zoneName: str = str(max(fnum+1, zoneNum))

                    # Keep extending the element connectivity until elemNodes is reached
                    _     = next(grsIter)
                    lnum += 1
                    zoneElems = []
                    while len(zoneElems) < zoneNElems:
                        lnum += 1
                        zoneElems.extend(next(grsIter).strip().split())

                    # Convert elemNodes to a numpy array of integers
                    zoneElems = np.array(zoneElems, dtype=np.uint64)

                    # Add the elem to the cellset
                    # > CS1: We create a dictionary of the zones and types that we want
                    for zoneElem in zoneElems:
                        cellsets[zoneName][elemTypes[zoneElem-1]].append(zoneElem-1)

                # Clean-up for memory safety
                del grsLine
                del grsIter

                # Search for the line starting the boundary conditions
                bcsLine  = lines_that_contain('BOUNDARY CONDITIONS', content)[0] + 1
                bcsIter  = iter(content[bcsLine:])
                lnum     = 0

                # Iterate and unpack the boundary conditions
                for line in bcsIter:
                    # Iterate until the number of boundary conditions is reached
                    lnum += 1
                    if 'ENDOFSECTION' in line:
                        # Check if the next sections is also a boundary condition
                        if bcsLine+lnum >= len(content) or 'BOUNDARY CONDITIONS' not in content[bcsLine+lnum]:
                            break

                    tokens = line.strip().split()
                    if not tokens:
                        continue

                    try:
                        bcName, bcType, bcnData, bcnVal, _ = tokens
                        bcName, bcType, bcnData, bcnVal    = bcName, int(bcType), int(bcnData), int(bcnVal)
                    except ValueError:
                        continue

                    # Ignore invalid (empty) boundary conditions
                    if bcnData <= 0:
                        continue

                    BCName = bcName.strip().lower()

                    match bcType:
                        # Nodal Data (ITYPE=0)
                        case 0:
                            raise NotImplementedError('Nodal data boundary conditions are not supported yet.')
                        # Element/Cell Data (ITYPE=1)
                        case 1:
                            # Read the next bcnData lines
                            bcnNodes = []
                            for _ in range(bcnData):
                                bcnNodes.extend(next(bcsIter).strip().split())
                                lnum += 1
                            # bcnNodes is in format [ELEM, ELEMTYPE, FACE, (VALUES)]
                            bcnNodes = np.array(bcnNodes, dtype=int).reshape(bcnData, -1)

                            # Attach the boundary sides
                            for elemID, gType, faceID in bcnNodes[:, :3]:
                                # Map gambit element type to meshio element type
                                elemType  = node_ordering.typing_gambit_to_meshio(gType)

                                # Get the face
                                elem      = cells[elemType][elemID-1]
                                face      = gambit_faces(elemType)[faceID-1]

                                # Determine the side name and number
                                nCorners  = len(face_to_cgns(face, elemType))
                                sideNum   = 0      if nCorners == 4 else 1           # noqa: E272
                                sideBase  = 'quad' if nCorners == 4 else 'triangle'  # noqa: E272
                                sideHO    = '' if mesh_vars.nGeo == 1 else str(NDOFperElemType(sideBase, mesh_vars.nGeo))
                                sideName  = sideBase + sideHO

                                # Map the face ordering from tensor-product to meshio
                                order     = FaceOrdering(sideBase, order=1)
                                corners   = elem[face_to_cgns(face, elemType)]
                                corners   = corners.flatten()[order]
                                sideNodes = np.expand_dims(corners, axis=0)

                                # Add the side to the cells
                                cells.setdefault(sideName, []).append(sideNodes.astype(np.uint64))

                                # Increment the side counter
                                nSides[sideNum] += 1

                                # Add the side to the cellset
                                # > CS1: We create a dictionary of the BC sides and types that we want
                                cellsets.setdefault(BCName, defaultdict()).setdefault(sideName, []).append(nSides[sideNum] - 1)

                # Clean-up for memory safety
                del bcsLine
                del bcsIter

    # After processing all elements, convert each list of arrays to one array
    # > Convert the list of cells to numpy arrays
    cells: dict = {cell_type: np.concatenate([a.reshape(1, -1) if a.ndim == 1 else a for a                      in cell_arrays])  # noqa: E272
                                                                                     for cell_type, cell_arrays in cells.items()}

    # Convert points_list back to a NumPy array
    points = np.array(pointl)

    # > CS2: We build the cell sets depending on the cells
    cell_sets:  dict[str, list] = mesh.cell_sets
    cell_types: list[Any      ] = list(cells.keys())
    nCellTypes: int             = len(cell_types)
    cell_tidx:  dict[Any, int ] = {ctype: idx for idx, ctype in enumerate(cell_types)}

    # Convert the dict of cellsets to numpy arrays
    for bc, bc_dict in cellsets.items():
        # Initialize entry for this BC if not exists
        if bc not in cell_sets:
            # Assign the entry to the cell set
            cell_sets[bc] = [None] * nCellTypes

        entry = cell_sets[bc]

        # Process all cell types for this BC
        for side, indices in bc_dict.items():
            BCIndices = np.fromiter(indices, dtype=np.uint64, count=len(indices))

            # Get cell type index
            type_idx = cell_tidx[side]

            # Find matching cell type and populate the corresponding entry
            if entry[type_idx] is not None:
                entry[type_idx] = np.concatenate([entry[type_idx], BCIndices])
            else:
                entry[type_idx] = BCIndices

    # > CS3: We create the final meshio.Mesh object with cell_sets
    mesh   = meshio.Mesh(points    = points,     # noqa: E251
                         cells     = cells,      # noqa: E251
                         cell_sets = cell_sets)  # noqa: E251

    # Run garbage collector to release memory
    gc.collect()

    return mesh



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import gc
import os
import re
import resource
import shutil
import subprocess
import time
from typing import Final, Optional, cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import h5py
import meshio
import numpy as np
from scipy.spatial import KDTree
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# Monkey-patching MeshIO
meshio._mesh.topological_dimension.update({'wedge15'   : 3,
                                           'pyramid13' : 3,
                                           'pyramid55' : 3})
# ==================================================================================================================================


def compatibleGMSH(file: str) -> bool:
    ioFormat = {1 : '.msh',
                2 : '.unv',
                # 10: 'auto',
                16: '.vtk',
                19: '.vrml',
                21: '.mail',
                # 26: 'pos stat',
                27: '.stl',
                28: '.p3d',
                30: '.mesh',
                31: '.bdf',
                32: '.cgns',
                33: '.med',
                34: '.diff',
                38: '.ir3',
                39: '.inp',
                40: '.ply2',
                41: '.celum',
                42: '.su2',
                47: '.tochnog',
                # 49: '.neu',   # Cubit/Gambit reader is broken beyond repair
                50: '.matlab'}

    # get file extension
    _, ext = os.path.splitext(file)
    return ext in ioFormat.values()


def ReadGMSH(fnames: list) -> meshio.Mesh:
    # Third-party libraries --------------------------------
    import gmsh
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.common.common import IsDisplay
    from pyhope.common.common_vars import np_mtp
    from pyhope.io.io_vars import debugvisu
    from pyhope.mesh.topology.mesh_serendipity import convertSerendipityToFullLagrange
    from pyhope.meshio.meshio_convert import gmsh_to_meshio
    from pyhope.readintools.readintools import GetLogical
    # ------------------------------------------------------

    # Setup stacksize
    resource.setrlimit(resource.RLIMIT_STACK, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))

    hopout.sep()
    gmsh.initialize()

    # Setup multiprocessing
    numThreads = np_mtp if np_mtp > 0 else 1
    gmsh.option.setNumber('General.NumThreads',   numThreads)
    gmsh.option.setNumber('Geometry.OCCParallel', 1 if np_mtp > 0 else 0)

    # Setup mesh factory
    # gmsh.option.setString('SetFactory', 'OpenCascade')

    # Setup debug visualization
    if not debugvisu:
        # Hide the GMSH debug output
        gmsh.option.setNumber('General.Terminal', 0)
        gmsh.option.setNumber('Geometry.Tolerance'         , 1e-12)  # default: 1e-6
        gmsh.option.setNumber('Geometry.MatchMeshTolerance', 1e-09)  # default: 1e-8

    for fname in fnames:
        # Get file extension
        _, ext = os.path.splitext(fname)

        gmsh.option.setNumber('Mesh.RecombineAll', 1)
        # gmsh.option.setNumber('Mesh.SecondOrderIncomplete', 0)

        # If not GMSH format convert
        if ext == '.cgns':
            # Setup GMSH to import required data
            # gmsh.option.setNumber('Mesh.SaveAll', 1)
            gmsh.option.setNumber('Mesh.CgnsImportIgnoreBC', 0)
            gmsh.option.setNumber('Mesh.CgnsImportIgnoreSolution', 1)

        # Enable agglomeration
        mesh_vars.already_curved = GetLogical('MeshIsAlreadyCurved')
        hopout.sep()
        if mesh_vars.already_curved and mesh_vars.nGeo > 1:
            if ext == '.cgns':
                gmsh.option.setNumber('Mesh.CgnsImportOrder', mesh_vars.nGeo)
            # Set the element order
            # > Technically, this is only required in generate_mesh but let's be precise here
            gmsh.model.mesh.setOrder(mesh_vars.nGeo)

        gmsh.merge(fname)

        # Explicitly load the OpenCASCADE kernel
        gmsh.model.occ.synchronize()

        # entities  = gmsh.model.getEntities()
        # nBCs_CGNS = len([s for s in entities if s[0] == 2])

        # Check if GMSH read all BCs
        # > This will only work if the CGNS file identifies elementary entities by CGNS "families" and by "BC" structures
        # > Possibly see upstream issue, https://gitlab.onelab.info/gmsh/gmsh/-/issues/2727\n'
        if ext == '.cgns':
            # WARNING: THIS PROBABLY NEVER WORKS, SO JUST USE OUR OWN APPROACH
            # if nBCs_CGNS == len(mesh_vars.bcs):
            #     for entDim, entTag in entities:
            #         # Surfaces are dim-1
            #         if entDim == 3:
            #             continue
            #
            #         entName = gmsh.model.get_entity_name(dim=entDim, tag=entTag)
            #         gmsh.model.addPhysicalGroup(entDim, [entTag], name=entName)
            # else:
            #     mesh_vars.CGNS.regenerate_BCs = True
            mesh_vars.CGNS.regenerate_BCs = True

        # gmsh.model.geo.synchronize()
        gmsh.model.occ.synchronize()

        # Optimize the high-order mesh
        # gmsh.model.mesh.optimize(method='Relocate3D', force=True)
        # gmsh.model.occ.synchronize()

    # Reclassify the nodes to ensure correct node ordering
    gmsh.model.mesh.reclassifyNodes()
    gmsh.model.occ.synchronize()

    if debugvisu and IsDisplay():
        gmsh.fltk.run()

    # Consistency check if the mesh contains volume elements
    # > User might have modified the mesh inside the FLTK GUI
    # gmshElems = np.asarray((gmsh.option.getNumber('Mesh.NbTetrahedra'),
    #                         gmsh.option.getNumber('Mesh.NbPrisms'    ),
    #                         gmsh.option.getNumber('Mesh.NbPyramids'  ),
    #                         gmsh.option.getNumber('Mesh.NbHexahedra')), dtype=int)
    gmshTypes = gmsh.model.mesh.getElementTypes()
    gmshElems = np.asarray([(elemName, order) for type                          in gmshTypes                                     # noqa: E272
                                               for elemName, dim, order, _, _, _ in [gmsh.model.mesh.getElementProperties(type)]  # noqa: E272
                              if dim == 3])
    if not np.any(gmshElems):
        hopout.error('Generated mesh does not contain volume elements, exiting...')

    # Consistency check if the mesh elements have the correct order
    gmshIssue  = np.asarray([(elemName, order) for type                          in gmshTypes                                     # noqa: E272
                                               for elemName, dim, order, _, _, _ in [gmsh.model.mesh.getElementProperties(type)]  # noqa: E272
                              if dim == 3 and order != mesh_vars.nGeo])

    if gmshIssue.size > 0:
        for elem in gmshIssue:
            print(hopout.warn(f'Wrong Gmsh order {elem[1]} for element {elem[0].replace(" ", "")}'))
        elemOrders = set([int(elem[1]) for elem in gmshIssue])
        hopout.error(f'Gmsh element order(s) {elemOrders} does not match requested mesh order {set([mesh_vars.nGeo])}')

    # Convert Gmsh object to meshio object
    mesh = gmsh_to_meshio(gmsh)

    # Check whether the mesh contains high-order elements and nGeo is set to 1
    if not mesh_vars.already_curved or mesh_vars.nGeo == 1:
        for elemtype in mesh.cells_dict.keys():
            if elemtype in mesh_vars.ELEMTYPE.name and mesh_vars.ELEMTYPE.name[elemtype] > 200:
                hopout.error('High-order elements detected in the mesh but MeshIsAlreadyCurved=F or nGeo is set to 1, exiting...')

    # If the mesh contains second-order incomplete elements, fix them
    mesh = convertSerendipityToFullLagrange(mesh)

    # Finally done with GMSH, finalize
    gmsh.finalize()

    # Convert BC names to lower case
    mesh.cell_sets = {k.lower(): v for k, v in mesh.cell_sets.items()}

    # Run garbage collector to release memory
    gc.collect()

    return mesh


def BCCGNS(mesh: meshio.Mesh, fnames: list) -> meshio.Mesh:
    """ Some CGNS files setup their boundary conditions in a different way than gmsh expects
        > Add them here manually to the meshIO object
    """
    # Standard libraries -----------------------------------
    import tempfile
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------

    hopout.routine('Applying boundary conditions')
    # hopout.sep()

    points  = mesh.points
    cells   = mesh.cells
    # elems   = mesh_vars.elems
    # sides   = mesh_vars.sides
    # bcs     = mesh_vars.bcs

    # All non-connected sides (technically all) are potential BC sides
    # nConnSide = [s for s in sides if 'Connection' not in s and 'BCID' not in s]
    cells_lst = tuple(mesh.cells_dict)

    # Now, for quadrilateral elements
    nConnLen = 0
    nConnNum = 0
    stree    = None

    if any('quad' in key for key in mesh.cells_dict):
        nConnSide = [value for key, value in mesh.cells_dict.items() if 'quad' in key][0]
        nConnType = [key   for key, _     in mesh.cells_dict.items() if 'quad' in key][0]  # noqa: E272, E501
        nConnNum  = cells_lst.index(nConnType)
        nConnLen  = len(cells_lst)

        # Collapse all opposing corner nodes into an [:, 12] array
        # nbCorners  = [s['Corners'] for s in nConnSide]
        nbCorners = [s[0:4] for s in nConnSide]
        # Calculate the centroid for each face (3D point)
        nbCenters = np.mean(mesh.points[nbCorners], axis=1)
        del nbCorners

        # Build a k-dimensional tree of all face centroids on the opposing side
        stree = KDTree(nbCenters, balanced_tree=False, compact_nodes=False)

    # Now, the same thing for triangular elements
    tConnLen  = 0
    tConnNum  = 0
    ttree     = None

    if any('triangle' in key for key in mesh.cells_dict):
        tConnSide = [value for key, value in mesh.cells_dict.items() if 'triangle' in key][0]
        tConnType = [key   for key, _     in mesh.cells_dict.items() if 'triangle' in key][0]  # FIXME: Support mixed LO/HO meshes  # noqa: E272, E501
        tConnNum  = cells_lst.index(tConnType)
        tConnLen  = len(cells_lst)

        # Collapse all opposing corner nodes into an [:, 9] array
        tbCorners = [s[0:3] for s in tConnSide]
        # Calculate the centroid for each face (3D point)
        tbCenters = np.mean(mesh.points[tbCorners], axis=1)
        del tbCorners

        # Build a k-dimensional tree of all face centroids
        ttree = KDTree(tbCenters, balanced_tree=False, compact_nodes=False)

    tol: Final[float] = mesh_vars.tolExternal

    # Now set the missing CGNS boundaries
    for fname in fnames:

        # Create a temporary directory and keep it existing until manually cleaned
        tfile = tempfile.NamedTemporaryFile(delete=False)
        tname = tfile.name
        # Try to convert the file automatically
        if not h5py.is_hdf5(fname):
            hopout.sep()
            hopout.info('File {} is not in HDF5 CGNS format, converting ...'.format(os.path.basename(fname)))
            tStart = time.time()
            _ = subprocess.run([f'adf2hdf {fname} {tname}'], check=True, shell=True, stdout=subprocess.DEVNULL)
            tEnd   = time.time()
            hopout.info('File {} converted HDF5 CGNS format [{:.2f} sec]'.format(os.path.basename(fname), tEnd - tStart))

            # Rest of this code operates on the converted file
            fname = tname
        else:
            # Alternatively, load the file directly into tmpfs for faster access
            shutil.copyfile(fname, tname)

        with h5py.File(fname, mode='r') as f:
            if 'CGNSLibraryVersion' not in f.keys():
                hopout.error('CGNS file does not contain library version header')

            key = [s for s in f.keys() if "base" in s.lower()]
            match len(key):
                case 0:
                    hopout.error('Object [Base] does not exist in CGNS file')
                case 1:
                    if not isinstance(f[key[0]], h5py.Group):
                        hopout.error('Object [Base] is not a group in CGNS file')
                    base = cast(h5py.Group, f[key[0]])
                case _:
                    hopout.error('More than one object [Base] exists in CGNS file')

            for baseZone in base.keys():
                # Ignore the base dataset
                if baseZone.strip() == 'data':
                    continue

                zone = cast(h5py.Group, base[baseZone])
                # Check if the zone contains BCs
                if 'ZoneBC' not in zone.keys():
                    continue

                zonedata = cast(h5py.Dataset, zone[' data'])
                match len(zonedata[0]):
                    case 1:  # Unstructured mesh, 1D arrays
                        mesh = BCCGNS_Unstructured(mesh, points, cells, stree, zone, tol, nConnNum, nConnLen,  # noqa: E501
                                                   # Support for triangular elements
                                                   ttree, tConnNum, tConnLen)
                    case 3:  # Structured 3D mesh, 3D arrays
                        # Structured grid can only contain tensor-product elements
                        mesh = BCCGNS_Structured(mesh, points, cells, stree, zone, tol, nConnNum, nConnLen)
                    case _:  # Unsupported number of dimensions
                        # raise ValueError('Unsupported number of dimensions')
                        hopout.error('Unsupported number of dimensions')

        # Cleanup temporary file
        if tfile is not None:
            os.unlink(tfile.name)

    # Run garbage collector to release memory
    gc.collect()
    hopout.sep()

    return mesh


def BCCGNS_Unstructured(  mesh:     meshio.Mesh,
                          points:   np.ndarray,
                          cells:    list,
                          stree:    Optional[KDTree],
                          zone,     # CGNS zone
                          tol:      float,
                          nConnNum: int,
                          nConnLen: int,
                          # Triangular elements
                          ttree:    Optional[KDTree],
                          tConnNum: int,
                          tConnLen: int) -> meshio.Mesh:
    """ Set the CGNS boundary conditions for uncurved (unstructured) grids
    """
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    from pyhope.io.formats.cgns import ElemTypes
    # ------------------------------------------------------
    # Load the CGNS points
    bpoints = np.column_stack([zone['GridCoordinates'][f'Coordinate{axis}'][' data'][:].astype(float) for axis in 'XYZ'])

    # Loop over all BCs
    zoneBCs  = [s for s in cast(h5py.Group, zone['ZoneBC']).keys() if s.strip() != 'innerfaces']
    cellsets = mesh.cell_sets
    # Convert the cellsets to a list of lists for easier manipulation
    for k, v in cellsets.items():
        cellsets[k] = list(map(lambda cell: cell.tolist() if isinstance(cell, (np.ndarray, np.generic)) else cell, v))

    for zoneBC in zoneBCs:
        # Lists to collect centroids
        quadCenters = []
        triaCenters = []

        # Data given with separate zoneBCs
        if zoneBC in zone:
            cgnsBC = cast(h5py.Dataset, zone[zoneBC]['ElementConnectivity'][' data'])

            # Read the surface elements, one at a time
            count  = 0

            # Loop over all elements and get the type
            while count < cgnsBC.shape[0]:
                elemType = ElemTypes(cgnsBC[count])

                # Map the unique quad sides to our non-unique elem sides
                nNodes   = int(elemType['Nodes'])
                corners  = cgnsBC[count+1:count+nNodes+1]
                # Calculate centroid from corner points and add to list
                quadCenters.append(np.mean(bpoints[corners - 1], axis=0))

                # Move to the next element
                count += nNodes + 1

        # Data attached to the zoneBC node
        elif f'{zoneBC}/PointList' in zone['ZoneBC']:
            cgnsBC = sorted(int(s.squeeze()) for s in zone['ZoneBC'][zoneBC]['PointList'][' data'])

            # Identify how surface elements are stored
            surface_key = 'GridShells' if 'GridShells' in zone else 'SurfaceElements' if 'SurfaceElements' in zone else None
            if not surface_key:
                hopout.error('Format of BC implementation for FaceCenters not recognized, exiting...')

            cgnsShells  =     zone[surface_key]['ElementConnectivity'][' data']
            nShells     = int(zone[surface_key]['ElementRange'][' data'][0])

            # Get the location of the BC faces
            cgnsGridLoc = bytes(zone['ZoneBC'][zoneBC]['GridLocation'][' data']).decode('ascii')
            cgns_set    = set(cgnsBC)

            # Read the surface elements, one at a time
            count   = 0

            # Loop over all elements and collect centroids
            while count < cgnsShells.shape[0]:
                elemType = ElemTypes(cgnsShells[count])
                nNodes   = int(elemType['Nodes'])

                corners  = cgnsShells[count+1:count+nNodes+1]

                if cgnsGridLoc == 'Vertex':
                    # Check if corners can form a subset of cgnsBC
                    corners_set = set(int(s) for s in corners)
                    if corners_set.issubset(cgns_set):
                        BCpoints = bpoints[[s-1 for s in corners]]
                        quadCenters.append(np.mean(BCpoints, axis=0))
                    count += nNodes + 1

                elif cgnsGridLoc == 'FaceCenter':
                    if nShells in cgnsBC:
                        BCpoints = bpoints[[s-1 for s in corners]]

                        # For high-order elements, we only consider the 3/4 corner nodes for the centroid
                        match len(BCpoints):
                            case 3 | 6:       # triangle, triangle6
                                triaCenters.append(np.mean(BCpoints[:3], axis=0))
                            case 4 | 8 | 9:   # quad, quad8, quad9
                                quadCenters.append(np.mean(BCpoints[:4], axis=0))
                            case _:
                                hopout.error('Unsupported number of corners for shell elements, exiting...')

                    nShells += 1
                    count   += nNodes + 1

        # Use regex to check if the string ends with _<number> and split accordingly
        match  = re.match(r'(.*)_\d+$', zoneBC)
        bcName = match.group(1) if match else zoneBC
        bcName = bcName.lower()

        # Process quads
        if quadCenters and stree is not None:
            distances, indices = cast(tuple[np.ndarray, np.ndarray], stree.query(np.array(quadCenters)))
            if np.any(distances > tol):
                hopout.error(f'Could not find all boundary sides within tolerance {tol} for BC "{bcName}", exiting...',
                             traceback=True)

            if bcName not in cellsets:
                cellsets[bcName] = [[] for _ in range(nConnLen)]
            cast(list, cellsets[bcName][nConnNum]).extend(indices.tolist())

        # Process triangles
        if triaCenters and ttree is not None:
            distances, indices = cast(tuple[np.ndarray, np.ndarray], ttree.query(np.array(triaCenters)))
            if np.any(distances > tol):
                hopout.error(f'Could not find all boundary sides within tolerance {tol} for BC "{bcName}", exiting...',
                             traceback=True)

            if bcName not in cellsets:
                cellsets[bcName] = [[] for _ in range(tConnLen)]
            cast(list, cellsets[bcName][tConnNum]).extend(indices.tolist())

    # Convert the cellsets back to a dictionary
    csets = {}
    for k, v in cellsets.items():
        csets[k] = [np.array(s, dtype=int) for s in v]

    mesh   = meshio.Mesh(points    = points,    # noqa: E251
                         cells     = cells,     # noqa: E251
                         cell_sets = csets)     # noqa: E251

    return mesh


def BCCGNS_Structured(mesh:     meshio.Mesh,
                      points:   np.ndarray,
                      cells:    list,
                      stree:    Optional[KDTree],
                      zone,     # CGNS zone
                      tol:      float,
                      nConnNum: int,
                      nConnLen: int) -> meshio.Mesh:
    """ Set the CGNS boundary conditions for (un)curved (structured) grids
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    # from pyhope.io.io_cgns import ElemTypes
    # ------------------------------------------------------
    # Loop over all BCs
    cellsets = mesh.cell_sets
    # Convert the cellsets to a list of lists for easier manipulation
    for k, v in cellsets.items():
        cellsets[k] = list(map(lambda cell: cell.tolist() if isinstance(cell, (np.ndarray, np.generic)) else cell, v))

    # Load the zone BCs
    for zoneBC, bcData in zone['ZoneBC'].items():
        try:
            cgnsBC   = bcData['FamilyName'][' data']
            cgnsName = ''.join(map(chr, cgnsBC)).lower()
        except KeyError:
            cgnsName = zoneBC.split('_', 1)[0].lower()

        # Ignore internal DEFAULT BCs
        if 'DEFAULT' in cgnsName:
            continue

        try:
            cgnsPointRange = np.array(bcData['PointRange'][' data'], dtype=int) - 1
            # Sanity check the CGNS point range
            if any(cgnsPointRange[1, :] - cgnsPointRange[0, :] < 0):
                hopout.error(f'Point range is not monotonically increasing on BC "{cgnsName}", exiting...')

            # Calculate the ranges of the indices
            iStart, iEnd = cgnsPointRange[:, 0]
            jStart, jEnd = cgnsPointRange[:, 1]
            kStart, kEnd = cgnsPointRange[:, 2]

            # Load the grid coordinates
            coords = {axis: np.array(zone['GridCoordinates'][f'Coordinate{axis}'][' data']) for axis in 'XYZ'}

            # Slice the grid
            slices = (slice(kStart, kEnd + 1), slice(jStart, jEnd + 1), slice(iStart, iEnd + 1))
            xSurf, ySurf, zSurf = (coords[axis][slices].squeeze() for axis in 'XYZ')

            # Dimensions of the surface grid
            iDim, jDim = xSurf.shape

            # Check if the grid dimensions can be sliced
            if (iDim - 1) % mesh_vars.nGeo != 0 or (jDim - 1) % mesh_vars.nGeo != 0:
                raise ValueError(f"Grid dimensions ({iDim}, {jDim}) are not divisible by the agglomeration factor {mesh_vars.nGeo}")

            # Slice the grid for agglomeration
            step = mesh_vars.nGeo
            xSurfNGeo, ySurfNGeo, zSurfNGeo = (arr[::step, ::step] for arr in (xSurf, ySurf, zSurf))

            # Updated dimensions after agglomeration
            iDimNGeo, jDimNGeo = xSurfNGeo.shape

            # Generate quads for the agglomerated grid
            # Define the quad by its four corner points
            quads = np.array([[(xSurfNGeo[j    , k    ], ySurfNGeo[j    , k    ], zSurfNGeo[j    , k    ]),
                               (xSurfNGeo[j + 1, k    ], ySurfNGeo[j + 1, k    ], zSurfNGeo[j + 1, k    ]),
                               (xSurfNGeo[j + 1, k + 1], ySurfNGeo[j + 1, k + 1], zSurfNGeo[j + 1, k + 1]),
                               (xSurfNGeo[j    , k + 1], ySurfNGeo[j    , k + 1], zSurfNGeo[j    , k + 1]) ] for j in range(iDimNGeo - 1)    # noqa: E501
                                                                                                             for k in range(jDimNGeo - 1)])  # noqa: E501

        except KeyError:
            hopout.error(f'ZoneBC "{zoneBC}" does not have a PointRange. PointLists are currently not supported.')

        if quads.size == 0:
            continue

        # Calculate centroids for all quads
        centers = np.mean(quads, axis=1)

        # Query the tree
        distances, indices = cast(tuple[np.ndarray, np.ndarray], stree.query(centers))
        if np.any(distances > tol):
            hopout.error(f'Could not find all boundary sides within tolerance {tol} for BC "{cgnsName}", exiting...',
                         traceback=True)

        # Update cellsets
        bcName = cgnsName.lower()
        if bcName not in cellsets:
            cellsets[bcName] = [[] for _ in range(nConnLen)]

        # Append all found side indices to the correct list
        cast(list, cellsets[bcName][nConnNum]).extend(indices.tolist())

    # Convert the cellsets back to a dictionary
    csets = {}
    for k, v in cellsets.items():
        csets[k] = [np.array(s, dtype=int) for s in v]

    mesh   = meshio.Mesh(points    = points,    # noqa: E251
                         cells     = cells,     # noqa: E251
                         cell_sets = csets)     # noqa: E251

    return mesh



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from functools import cache
# from typing import Tuple
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


@cache
def ElemTypes(num: int) -> dict[str, str | int]:
    types = [{ 'ElemTypeCGNS': 'ElementTypeNull', 'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 0
             { 'ElemTypeCGNS': 'ElementTypeUser', 'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 1
             { 'ElemTypeCGNS': 'NODE',            'ElemTypeMeshIO': 'vertex',       'Nodes': 1  },  # 2
             { 'ElemTypeCGNS': 'BAR_2',           'ElemTypeMeshIO': 'line',         'Nodes': 2  },  # 3
             { 'ElemTypeCGNS': 'BAR_3',           'ElemTypeMeshIO': 'line3',        'Nodes': 3  },  # 4
             { 'ElemTypeCGNS': 'TRI_3',           'ElemTypeMeshIO': 'triangle',     'Nodes': 3  },  # 5
             { 'ElemTypeCGNS': 'TRI_6',           'ElemTypeMeshIO': 'triangle6',    'Nodes': 6  },  # 6
             { 'ElemTypeCGNS': 'QUAD_4',          'ElemTypeMeshIO': 'quad',         'Nodes': 4  },  # 7
             { 'ElemTypeCGNS': 'QUAD_8',          'ElemTypeMeshIO': 'quad8',        'Nodes': 8  },  # 8
             { 'ElemTypeCGNS': 'QUAD_9',          'ElemTypeMeshIO': 'quad9',        'Nodes': 9  },  # 9
             { 'ElemTypeCGNS': 'TETRA_4',         'ElemTypeMeshIO': 'tetra',        'Nodes': 4  },  # 10
             { 'ElemTypeCGNS': 'TETRA_10',        'ElemTypeMeshIO': 'tetra10',      'Nodes': 10 },  # 11
             { 'ElemTypeCGNS': 'PYRA_5',          'ElemTypeMeshIO': 'pyramid',      'Nodes': 5  },  # 12
             { 'ElemTypeCGNS': 'PYRA_14',         'ElemTypeMeshIO': 'pyramid14',    'Nodes': 14 },  # 13
             { 'ElemTypeCGNS': 'PENTA_6',         'ElemTypeMeshIO': 'wedge',        'Nodes': 6  },  # 14
             { 'ElemTypeCGNS': 'PENTA_15',        'ElemTypeMeshIO': 'wedge15',      'Nodes': 15 },  # 15
             { 'ElemTypeCGNS': 'PENTA_18',        'ElemTypeMeshIO': 'wedge18',      'Nodes': 18 },  # 16
             { 'ElemTypeCGNS': 'HEXA_8',          'ElemTypeMeshIO': 'hexahedron',   'Nodes': 8  },  # 17
             { 'ElemTypeCGNS': 'HEXA_20',         'ElemTypeMeshIO': 'hexahedron20', 'Nodes': 20 },  # 18
             { 'ElemTypeCGNS': 'HEXA_27',         'ElemTypeMeshIO': 'hexahedron27', 'Nodes': 27 },  # 19
             { 'ElemTypeCGNS': 'MIXED',           'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 20
             { 'ElemTypeCGNS': 'PYRA_13',         'ElemTypeMeshIO': 'pyramid13',    'Nodes': 13 },  # 21
             { 'ElemTypeCGNS': 'NGON_n',          'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 22
             { 'ElemTypeCGNS': 'NFACE_n',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 23
             { 'ElemTypeCGNS': 'BAR_4',           'ElemTypeMeshIO': 'line4',        'Nodes': 4  },  # 24
             { 'ElemTypeCGNS': 'TRI_9',           'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 25
             { 'ElemTypeCGNS': 'TRI_10',          'ElemTypeMeshIO': 'triangle10',   'Nodes': 10 },  # 26
             { 'ElemTypeCGNS': 'QUAD_12',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 27
             { 'ElemTypeCGNS': 'QUAD_16',         'ElemTypeMeshIO': 'quad16',       'Nodes': 16 },  # 28
             { 'ElemTypeCGNS': 'TETRA_16',        'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 29
             { 'ElemTypeCGNS': 'TETRA_20',        'ElemTypeMeshIO': 'tetra20',      'Nodes': 20 },  # 30
             { 'ElemTypeCGNS': 'PYRA_21',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 31
             { 'ElemTypeCGNS': 'PYRA_29',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 32
             { 'ElemTypeCGNS': 'PYRA_30',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 33
             { 'ElemTypeCGNS': 'PENTA_24',        'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 34
             { 'ElemTypeCGNS': 'PENTA_38',        'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 35
             { 'ElemTypeCGNS': 'PENTA_40',        'ElemTypeMeshIO': 'wedge40',      'Nodes': 40 },  # 36
             { 'ElemTypeCGNS': 'HEXA_32',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 37
             { 'ElemTypeCGNS': 'HEXA_56',         'ElemTypeMeshIO': 'Null',         'Nodes': 0  },  # 38
             { 'ElemTypeCGNS': 'HEXA_64',         'ElemTypeMeshIO': 'hexahedron64', 'Nodes': 64 },  # 39
             { 'ElemTypeCGNS': 'Null',            'ElemTypeMeshIO': 'hexahedron24', 'Nodes': 24 },  # 40
             ]
    return types[num]


# def spiral_matrix(n: int) -> np.ndarray:
#     """ Print a spiral matrix of order n
#         > https://rosettacode.org/wiki/Spiral_matrix#Simple_solution
#     """
#     m = [[0] * n for _ in range(n)]
#     dx, dy  = [0, 1, 0, -1], [1, 0, -1, 0]
#     x, y, c =  0, -1, 1
#     for i in range(n + n - 1):
#         for j in range((n + n - i) // 2):
#             x += dx[i % 4]
#             y += dy[i % 4]
#             m[x][y] = c
#             c += 1
#     return np.array(m)
#
#
# def edgePointCGNS(order: int, edge: int, node: int) -> np.ndarray:
#     match edge:
#         case 0:  # z- / base
#             return np.array([node      , 0         ], dtype=int)
#         case 1:  # y+ / base
#             return np.array([order     , node      ], dtype=int)
#         case 2:  # z+ / base
#             return np.array([order-node, order     ], dtype=int)
#         case 3:  # y- / base
#             return np.array([0         , order-node], dtype=int)
#         case _:
#             sys.exit(1)
#
#
# @cache
# def HEXMAPCGNS(order: int) -> Tuple[np.ndarray, np.ndarray]:
#     """ CGNS -> IJK ordering for high-order hexahedrons
#         > Loosely based on [Gmsh] "generatePointsHexCGNS"
#
#         > HEXTEN : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (1D, tensor-product style)
#         > HEXMAP : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (3D mapping)
#     """
#     map = np.zeros((order, order, order), dtype=int)
#
#     if order == 1:
#         map[0, 0, 0] = 0
#         tensor       = map
#         return map, tensor
#
#     # Principal vertices
#     map[0      , 0      , 0      ] = 1
#     map[order-1, 0      , 0      ] = 2
#     map[order-1, order-1, 0      ] = 3
#     map[0      , order-1, 0      ] = 4
#     map[0      , 0      , order-1] = 5
#     map[order-1, 0      , order-1] = 6
#     map[order-1, order-1, order-1] = 7
#     map[0      , order-1, order-1] = 8
#
#     if order == 2:
#         # Python indexing, 1 -> 0
#         map -= 1
#         # Reshape into 1D array, tensor-product style
#         tensor = []
#         for k in range(order):
#             for j in range(order):
#                 for i in range(order):
#                     tensor.append(int(map[i, j, k]))
#
#         return map, np.asarray(tensor)
#
#     # Internal points of base quadrangle edges (x-)
#     count = 8
#     for iFace in range(4):
#         for iNode in range(1, order-1):
#             # Assemble mapping to tuple, base quadrangle -> z = 0
#             count += 1
#             edge  = edgePointCGNS(order-1, iFace, iNode)
#             index = (int(edge[0]), int(edge[1]), 0)
#             map[index] = count
#
#     # Internal points of mounting edges
#     for iFace in range(4):
#         edge  = edgePointCGNS(order-1, (iFace+3) % 4, order-1)
#         for iNode in range(1, order-1):
#             # Assemble mapping to tuple, mounting edges -> z ascending
#             count += 1
#             index = (int(edge[0]), int(edge[1]), iNode)
#             map[index] = count
#
#     # Internal points of top quadrangle edges
#     for iFace in range(4):
#         for iNode in range(1, order-1):
#             # Assemble mapping to tuple, top  quadrangle -> z = order
#             count += 1
#             edge  = edgePointCGNS(order-1, iFace, iNode)
#             index = (int(edge[0]), int(edge[1]), order-1)
#             map[index] = count
#
#     # Internal points of base quadrangle
#     k    = 0
#     # Fill in spirals
#     face = spiral_matrix(order-2)
#     # Fill the map
#     for j in range(order-2):
#         for i in range(order-2):
#             index = (i+1  , j+1  , k    )
#             map[index] = count + face[j, i]
#     count += (order-2)**2
#
#     # Internal points of upstanding faces
#     # > y- face
#     k    = 0
#     # Fill the map
#     for j in range(order-2):
#         for i in range(order-2):
#             index = (i+1  , k    , j+1  )
#             map[index] = count + face[j, i]
#     count += (order-2)**2
#
#     # > x+ face
#     k    = order-1
#     for j in range(order-2):
#         for i in range(order-2):
#             index = (k    , i+1  , j+1  )
#             map[index] = count + face[j, i]
#     count += (order-2)**2
#
#     # > y+ face
#     k      = order-1
#     face_r = np.rot90(face, k=2, axes=(1, 0))
#     for j in range(order-2):
#         for i in range(order-2):
#             index = (i+1  , k    , j+1  )
#             map[index] = count + face_r[order-3-j, i]
#     count += (order-2)**2
#
#     # > x- face
#     k      = 0
#     for j in range(order-2):
#         for i in range(order-2):
#             index = (k    , i+1  , j+1  )
#             map[index] = count + face_r[order-3-j, i]
#     count += (order-2)**2
#
#     # Internal points of top  quadrangle
#     k    = order-1
#     # Fill in spirals
#     face = spiral_matrix(order-2)
#     # Fill the map
#     for j in range(order-2):
#         for i in range(order-2):
#             index = (i+1  , j+1  , k    )
#             map[index] = count + face[j, i]
#     count += (order-2)**2
#
#     # Internal volume points as a succession of internal planes
#     for k in range(1, order-1):
#         for j in range(order-2):
#             for i in range(order-2):
#                 index = (i+1  , j+1  , k    )
#                 map[index] = count + face[j, i]
#         count += (order-2)**2
#
#     # Python indexing, 1 -> 0
#     map -= 1
#
#     # Reshape into 1D array, tensor-product style
#     tensor = []
#     for k in range(order):
#         for j in range(order):
#             for i in range(order):
#                 tensor.append(int(map[i, j, k]))
#
#     return map, np.asarray(tensor)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import json
import urllib.request
from packaging.requirements import Requirement
from packaging.version import Version
from typing import Optional, Union
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def PyPIVersion(package: str, timeout: int = 10) -> Optional[Version]:
    """ Query PyPI JSON API for the latest version

        Returns:
            Optional[str]: Version string
    """
    url = f"https://pypi.org/pypi/{package}/json"
    try:
        with urllib.request.urlopen(url, timeout=timeout) as resp:
            if resp.status != 200:
                return None
            data = json.load(resp)
            info = data.get('info', {})
            return Version(info.get('version'))
    except Exception:  # pragma: no cover
        return None


def _ParseVersion(text: str) -> Optional[Version]:
    """ Find a semver-like version in arbitrary text
    """
    # Local imports ----------------------------------------
    from re import search
    from packaging.version import Version
    # ------------------------------------------------------
    if not text:
        return None

    # 1) Prefer to capture numeric version before a "-git" marker
    #    Example: "4.14.0-git-8425b99f0" -> capture "4.14.0"
    m = search(r"\bv?(\d+(?:\.\d+){0,3})(?=-git\b)", text)
    if m:
        try:
            return Version(m.group(1))
        except Exception:  # pragma: no cover
            return None

    # 2) Fallback: capture semver-like token including -rc / +build etc.
    #    Example: "v1.2.3-rc1", "1.2.3+build.1"
    m = search(r"\bv?(\d+(?:\.\d+){0,3}(?:[-_a-zA-Z0-9+.]+)?)\b", text)
    if m:
        try:
            return Version(m.group(1))
        except Exception:  # pragma: no cover
            return None

    return None


def GmshVersion() -> tuple[Union[Version, bool, None], Union[str, None]]:
    """ Query the local system for the Gmsh version

        Returns:
            Optional[str]: Version string
    """
    # Local imports ----------------------------------------
    import re
    import shutil
    import subprocess
    # ------------------------------------------------------
    path = shutil.which('gmsh')
    ver: Union[Version, bool, None] = None
    pac: Union[str,           None] = None  # noqa: E272

    if path:
        try:
            p   = subprocess.run([path, '--info'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, timeout=5)
            raw = (p.stdout or '') + "\n" + (p.stderr or '')
            # Parse version from the Version line if present
            v   = _ParseVersion(raw)
            if v:
                ver = v
            # Parse Packaged by: <who>
            m   = re.search(r'Packaged by\s*:\s*(.+)', raw, flags=re.IGNORECASE)
            if m:
                pac = str(m.group(1).strip())
        except Exception:  # pragma: no cover
            pass

    # If no version found yet, try generic flags
    ver = DependencyVersion('gmsh')

    # Finalize fields and return
    if not ver:  # pragma: no cover
        ver = None
    if not pac:  # pragma: no cover
        pac = None
    return ver, pac


def DependencyVersion(program: str) -> Optional[Version | bool]:
    """ Query the local system for the version of a dependency

        Returns:
            Optional[str]: Version string
    """
    # Local imports ----------------------------------------
    import shutil
    import subprocess
    # ------------------------------------------------------
    path = shutil.which(program)

    # Try common version flags. Some executables print version to stdout, others to stderr
    if path:
        for flag in ('--version', '-V', 'version', '-v'):
            try:
                p = subprocess.run([path, flag], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, timeout=5)
            except Exception:
                continue

            out = (p.stdout or '') + "\n" + (p.stderr or '')
            ver = _ParseVersion(out)
            if ver:
                return ver

        # Found dependency but failed to parse version
        return True
    return None


def DependencyHealth(program: str,
                     version: Union[Version, bool, None],
                     status:  Optional[str] = None,
                     info:    Optional[str] = None) -> None:
    """ Print the dependency health
    """
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    # ------------------------------------------------------

    if version:
        if isinstance(version, Version):
            status = status if status is not None else hopout.Symbols.OK
            hopout.info(f'{status} {program} found [v{version}]'      + ('' if info is None else info))
        else:  # pragma: no cover
            status = status if status is not None else hopout.Symbols.WARN
            hopout.info(f'{status} {program} found [unknown version]' + ('' if info is None else info))
    else:
        status = status if status is not None else hopout.Symbols.WARN
        hopout.info(f'{status} {program} not installed' + ('' if info is None else info))


def _PackageInstalledVersion(package: str) -> Optional["Version"]:
    """ Query the local system for the version of a Python package

        Returns:
            Optional[str]: Version string
    """
    # Local imports ----------------------------------------
    import importlib
    from importlib import metadata as importlib_metadata
    from packaging.version import Version
    # ------------------------------------------------------
    # Try distribution metadata first (package name as on PyPI)
    try:
        v = importlib_metadata.version(package)
        return Version(v)
    except Exception:
        pass

    # Fallback: try importing top-level module and reading __version__
    try:
        mod = importlib.import_module(package)
        v = getattr(mod, '__version__', None) or getattr(mod, 'version', None) or getattr(mod, 'release', None)
        if v:
            try:
                return Version(str(v))
            except Exception:  # pragma: no cover
                return None
    except Exception:  # pragma: no cover
        pass

    return None


def PackageHealth(   pkg:    str,
                     version: Union[Version, None],
                     pypiver: Union[Version, None]  # noqa: E272
                 ) -> None:
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    # ------------------------------------------------------
    if pypiver and version:
        try:
            if version >= pypiver:
                hopout.info(f'{hopout.Symbols.OK  } {pkg} [v{version}] is up-to-date')
            else:
                hopout.info(f'{hopout.Symbols.WARN} {pkg} [v{version}] is outdated (PyPI: v{pypiver})')
        except Exception:  # pragma: no cover
            hopout.info(f'{hopout.Symbols.WARN} {pkg} [v{version}] is installed (PyPI: v{pypiver}) -- unable to compare reliably')
    elif version:  # pragma: no cover
        hopout.info(f'{hopout.Symbols.WARN} {pkg} [v{version}] is installed (PyPI info unavailable)')
    else:  # pragma: no cover
        hopout.info(f'{hopout.Symbols.ERR} {pkg} [v{version}] is not installed')


def CheckHealth() -> None:
    """ Internal health check
    """
    # Local imports ----------------------------------------
    from pyhope.common.common_vars import Common
    import pyhope.output.output as hopout
    from importlib import metadata as importlib_metadata
    # ------------------------------------------------------

    common  = Common()
    program = common.program
    symbols = hopout.Symbols

    hopout.small_banner('System')

    pypiver = PyPIVersion('pyhope')
    PackageHealth(common.program, Version(common.version), pypiver)

    hopout.info('')
    hopout.small_banner('Packages')

    # Try to read the installed distribution metadata
    try:
        dist = importlib_metadata.distribution(program)
        pkgs = dist.requires or []
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        # Fallback: try metadata() api to read Requires-Dist directly
        try:
            meta = importlib_metadata.metadata(program)
            # metadata.get_all returns a sequence of Requires-Dist entries if present
            pkgs = list(meta.get_all('Requires-Dist') or [])
        except Exception:
            pkgs = []

    if not pkgs:  # pragma: no cover
        print(hopout.warn(f'Could not discover declared Python requirements for {program}.'))
        print(hopout.warn( 'If you are running from a source checkout, install the package first (pip install -e .) to ' +
                           'enable requirement checks.'))

    # For optional dependencies, split the first part
    pkgs = set(p.split(';')[0] if len(p.split(';')) > 0 else p for p in pkgs)
    for pack in sorted(pkgs):
        # packaging.Requirement will handle extras and environment markers
        try:
            req = Requirement(pack)
        except Exception:  # pragma: no cover
            # If parsing fails, fall back to a simple split at first space or semicolon
            raw_name = pack.split()[0].split(';')[0]
            name = raw_name.split('(')[0].strip()
            req  = None
            pkg  = name
        else:  # pragma: no cover
            pkg  = req.name

        pkgver  = _PackageInstalledVersion(pkg)
        pypiver = PyPIVersion(pkg)
        PackageHealth(pkg, pkgver, pypiver)

    hopout.info('')
    hopout.small_banner('Dependencies')

    # For Gmsh, also check the packager
    gmshv, gmshp = GmshVersion()
    if gmshp is not None and gmshp.strip() == 'NRG':
        gmshi = ' (packaged by NRG)'
        gmshs = symbols.OK
    else:  # pragma: no cover
        gmshi = ' (packaged externally)'
        gmshs = symbols.WARN

    DependencyHealth('Gmsh'    , version=gmshv,           status=gmshs, info=gmshi)
    # Warn if we know that Gmsh uses an outdated CGNS
    if gmshp is not None and gmshp.strip() != 'NRG':  # pragma: no cover
        print(hopout.warn('Detected Gmsh package uses an outdated CGNS (v3.4). ' +
                          'For compatibility, replace with the updated NRG version'))
    DependencyHealth('ParaView', version=DependencyVersion('paraview'), info=' (optional)')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from functools import cache
from typing import Tuple
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


@cache                                                       # pragma: no cover
def HEXMAPVTK(order: int) -> Tuple[np.ndarray, np.ndarray]:  # pragma: no cover
    """ VTK -> IJK ordering for high-order hexahedrons
        > Loosely based on [Gmsh] "generatePointsHexCGNS"
        > [Jens Ulrich Kreber] "paraview-scripts/node_ordering. py"

        > The following code is actually based on FLEXI MOD_VTK and trixi. jl.  Note from trixi:
        > > This order doesn't make any sense.  This is completely different from what is shown in
        > > https://blog.kitware.com/wp-content/uploads/2018/09/Source_Issue_43.pdf but this is the way it works.

        > HEXTEN :  np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (1D, tensor-product style)
        > HEXMAP : np.ndarray # MESHIO <-> IJK ordering for high-order hexahedrons (3D mapping)
    """
    if order == 1:
        map          = np.zeros((1, 1, 1), dtype=int)
        map[0, 0, 0] = 0
        tensor       = np.array([0])
        return map, tensor

    # Pre-allocate map array
    map = np.zeros((order, order, order), dtype=int)
    idx = 0

    # Principal vertices
    map[0      , 0      , 0      ] = 0
    map[order-1, 0      , 0      ] = 1
    map[order-1, order-1, 0      ] = 2
    map[0      , order-1, 0      ] = 3
    map[0      , 0      , order-1] = 4
    map[order-1, 0      , order-1] = 5
    map[order-1, order-1, order-1] = 6
    map[0      , order-1, order-1] = 7
    idx += 8

    if order > 2:
        # Number of interior nodes per edge
        nEdge = order - 2
        inner = np.arange(1, order-1)

        # Internal points of mounting edges
        map[inner  , 0      , 0      ] = np.arange(idx, idx+nEdge)  # Edge 1:  x-edge, y=0  , z=0
        idx += nEdge
        map[order-1, inner  , 0      ] = np.arange(idx, idx+nEdge)  # Edge 2:  y-edge, x=max, z=0
        idx += nEdge
        map[inner  , order-1, 0      ] = np.arange(idx, idx+nEdge)  # Edge 3:  x-edge, y=max, z=0
        idx += nEdge
        map[0      , inner  , 0      ] = np.arange(idx, idx+nEdge)  # Edge 4:  y-edge, x=0  , z=0
        idx += nEdge
        map[inner  , 0      , order-1] = np.arange(idx, idx+nEdge)  # Edge 5:  x-edge, y=0  , z=max
        idx += nEdge
        map[order-1, inner  , order-1] = np.arange(idx, idx+nEdge)  # Edge 6:  y-edge, x=max, z=max
        idx += nEdge
        map[inner  , order-1, order-1] = np.arange(idx, idx+nEdge)  # Edge 7:  x-edge, y=max, z=max
        idx += nEdge
        map[0      , inner  , order-1] = np.arange(idx, idx+nEdge)  # Edge 8:  y-edge, x=0  , z=max
        idx += nEdge
        map[0      , 0      , inner  ] = np.arange(idx, idx+nEdge)  # Edge 9:  z-edge, x=0  , y=0
        idx += nEdge
        map[order-1, 0      , inner  ] = np.arange(idx, idx+nEdge)  # Edge 10: z-edge, x=max, y=0
        idx += nEdge
        # INFO: The following two are switched compared to trixi because ParaView changed the ordering from VTK8 to VTK9 convention
        # https://gitlab.kitware.com/paraview/paraview/-/issues/20728
        map[order-1, order-1, inner  ] = np.arange(idx, idx+nEdge)  # Edge 11: z-edge, x=max, y=max
        idx += nEdge
        map[0      , order-1, inner  ] = np.arange(idx, idx+nEdge)  # Edge 12: z-edge, x=0  , y=max
        idx += nEdge

        # Number of interior nodes per face
        nFace = nEdge * nEdge

        # Faces:
        # > First index varies fastest, so emulate Fortran column-major via 'C' ravel on 'xy' meshgrid
        ii, jj = np.meshgrid(inner, inner, indexing='xy')
        iFace  = ii.ravel(order='C')
        jFace  = jj.ravel(order='C')
        del ii, jj

        map[0      , iFace, jFace] = np.arange(idx, idx+nFace)  # Face 1: x=0   (Left)   - yz plane
        idx += nFace
        map[order-1, iFace, jFace] = np.arange(idx, idx+nFace)  # Face 2: x=max (Right)  - yz plane
        idx += nFace
        map[iFace, 0      , jFace] = np.arange(idx, idx+nFace)  # Face 3: y=0   (Front)  - xz plane
        idx += nFace
        map[iFace, order-1, jFace] = np.arange(idx, idx+nFace)  # Face 4: y=max (Back)   - xz plane
        idx += nFace
        map[iFace, jFace, 0      ] = np.arange(idx, idx+nFace)  # Face 5: z=0   (Bottom) - xy plane
        idx += nFace
        map[iFace, jFace, order-1] = np.arange(idx, idx+nFace)  # Face 6: z=max (Top)    - xy plane
        idx += nFace

        # Volume
        # > Fortran order: i varies fastest, then j, then k
        nVol = nEdge ** 3
        ii, jj, kk = np.meshgrid(inner, inner, inner, indexing='ij')
        iVol = ii.ravel(order='F')
        jVol = jj.ravel(order='F')
        kVol = kk.ravel(order='F')
        del ii, jj, kk

        map[iVol, jVol, kVol] = np.arange(idx, idx+nVol)
        idx += nVol

    # Reshape into 1D array, tensor-product style
    tensor = np.zeros(order ** 3, dtype=int)
    for k in range(order):
        for j in range(order):
            for i in range(order):
                tensor[map[i, j, k]] = i + j * order + k * order * order

    return map, tensor



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.


# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from collections import namedtuple
from contextlib import contextmanager
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
from pyhope.basis.basis_basis import legendre_gauss_nodes, legendre_gauss_lobatto_nodes
from pyhope.basis.basis_basis import barycentric_weights, polynomial_derivative_matrix
from pyhope.basis.basis_basis import lagrange_interpolation_polys, calc_vandermonde
from pyhope.basis.basis_basis import change_basis_3D, change_basis_2D
from pyhope.basis.basis_basis import evaluate_jacobian
# ==================================================================================================================================


class Basis:
    """ Basis class to hold all basis related functions and variables
    """
    legendre_gauss_nodes         = staticmethod(legendre_gauss_nodes)
    legendre_gauss_lobatto_nodes = staticmethod(legendre_gauss_lobatto_nodes)
    barycentric_weights          = staticmethod(barycentric_weights)
    polynomial_derivative_matrix = staticmethod(polynomial_derivative_matrix)
    lagrange_interpolation_polys = staticmethod(lagrange_interpolation_polys)
    calc_vandermonde             = staticmethod(calc_vandermonde)
    change_basis_3D              = staticmethod(change_basis_3D)
    change_basis_2D              = staticmethod(change_basis_2D)
    evaluate_jacobian            = staticmethod(evaluate_jacobian)


# Define a named tuple to hold the mesh data
MeshContainer = namedtuple('Mesh',
                          ['mesh',   # The generated mesh object
                           'nGeo',   # Polynomial order
                           'bcs',    # Boundary conditions
                           'elems',  # Elements
                           'sides'   # Sides
                          ])


@contextmanager  # pragma: no cover
def Mesh(*args: str, stdout: bool = False, stderr: bool = True):
    """ Mesh context manager to generate a mesh from a given file

        Args:
            *args: The mesh file path(s) to be processed
            stdout (bool): If False, standard output is suppressed
            stderr (bool): If False, standard error  is suppressed

        Yields:
            Mesh: An object containing the generated mesh and its properties
    """
    # Standard libraries -----------------------------------
    import os
    # Third-party libraries --------------------------------
    import h5py
    from contextlib import redirect_stdout, redirect_stderr, ExitStack
    # Local imports ----------------------------------------
    import pyhope.config.config as config
    import pyhope.mesh.mesh_vars as mesh_vars
    from pyhope.common.common import DefineCommon, InitCommon
    from pyhope.io.io import DefineIO, InitIO
    from pyhope.mesh.connect.connect import ConnectMesh
    from pyhope.mesh.mesh import DefineMesh, InitMesh, GenerateMesh
    from pyhope.mesh.mesh_sides import GenerateSides
    from pyhope.readintools.readintools import DefineConfig, ReadConfig
    # ------------------------------------------------------

    try:
        # Check if the arguments provided are valid mesh files
        if not args:
            raise ValueError('No mesh file provided.')

        for arg in args:
            # Check if the argument is a valid file path
            if not os.path.isfile(arg):
                raise FileNotFoundError(f'Mesh file not found: {arg}')

            # Check if the argument is a valid HDF5 file
            if not h5py.is_hdf5(arg):
                raise ValueError(f'Mesh file not a valid HDF5 file: {arg}')

        # Suppress output to standard output
        with ExitStack() as stack:
            with open(os.devnull, 'w') as null:
                if not stdout:
                    stack.enter_context(redirect_stdout(null))
                if not stderr:
                    stack.enter_context(redirect_stderr(null))

                # Perform the reduced PyHOPE initialization
                with DefineConfig() as dc:
                    config.prms = dc
                    DefineCommon()
                    DefineIO()
                    DefineMesh()

                with ReadConfig(args[0]) as rc:
                    config.params = rc

                # Read-in required parameters
                InitCommon()
                InitIO()
                InitMesh()

                # Generate the actual mesh
                GenerateMesh()

                # Build our data structures
                GenerateSides()
                ConnectMesh()

        # Export mesh variables
        mesh = mesh_vars.mesh

        nGeo  = mesh_vars.nGeo
        bcs   = mesh_vars.bcs

        elems = mesh_vars.elems
        sides = mesh_vars.sides

        yield MeshContainer(mesh  = mesh,   # noqa: E251
                            nGeo  = nGeo,   # noqa: E251
                            bcs   = bcs,    # noqa: E251
                            elems = elems,  # noqa: E251
                            sides = sides   # noqa: E251
                           )

    finally:
        # Cleanup resources after exiting the context
        del mesh_vars.mesh
        del mesh_vars.nGeo
        del mesh_vars.bcs
        del mesh_vars.elems
        del mesh_vars.sides



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import copy
import gc
import math
import resource
from typing import cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import meshio
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def MeshCartesian() -> meshio.Mesh:
    # Third-party libraries --------------------------------
    import gmsh
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.common.common import find_index, find_indices, IsDisplay
    from pyhope.common.common_vars import np_mtp
    from pyhope.io.io_vars import debugvisu
    from pyhope.mesh.mesh_common import edge_to_dir, face_to_corner, face_to_edge, faces
    from pyhope.mesh.mesh_vars import BC
    from pyhope.mesh.transform.mesh_transform import CalcStretching
    from pyhope.meshio.meshio_convert import gmsh_to_meshio
    from pyhope.readintools.readintools import CountOption, GetInt, GetIntArray, GetIntFromStr, GetRealArray, GetStr
    # ------------------------------------------------------

    # Setup stacksize
    resource.setrlimit(resource.RLIMIT_STACK, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))

    gmsh.initialize()

    # Setup multiprocessing
    numThreads = np_mtp if np_mtp > 0 else 1
    gmsh.option.setNumber('General.NumThreads',   numThreads)
    gmsh.option.setNumber('Geometry.OCCParallel', 1 if np_mtp > 0 else 0)

    # Setup mesh factory
    # gmsh.option.setString('SetFactory', 'OpenCascade')

    # Setup debug visualization
    if not debugvisu:
        # Hide the GMSH debug output
        gmsh.option.setNumber('General.Terminal', 0)
        gmsh.option.setNumber('Geometry.Tolerance'         , 1e-12)  # default: 1e-6
        gmsh.option.setNumber('Geometry.MatchMeshTolerance', 1e-09)  # default: 1e-8

    hopout.sep()

    nZones    = GetInt('nZones')
    elemTypes = [int() for _ in range(nZones)]

    offsetp   = 0
    offsets   = 0

    # GMSH only supports mesh elements within a single model
    # > https://gitlab.onelab.info/gmsh/gmsh/-/issues/2836
    gmsh.model.add('Domain')
    gmsh.model.set_current('Domain')
    bcZones = [list() for _ in range(nZones)]

    for zone in range(nZones):
        hopout.routine('Generating zone {}'.format(zone+1))

        # check if corners are given in the input file
        if CountOption('Corner') > 0:
            corners  = GetRealArray( 'Corner'  , number=zone)
        elif CountOption('DX') > 0:
            # get extension of the computational zone
            DX = GetRealArray( 'DX'  , number=zone)

            # read in origin of the zone
            X0 = GetRealArray( 'X0'  , number=zone)

            # reconstruct points from DX and X0 such that all corners are defined
            corners = np.array((np.array((X0[0],       X0[1],       X0[2]      )),
                                np.array((X0[0]+DX[0], X0[1],       X0[2]      )),
                                np.array((X0[0]+DX[0], X0[1]+DX[1], X0[2]      )),
                                np.array((X0[0],       X0[1]+DX[1], X0[2]      )),
                                np.array((X0[0],       X0[1],       X0[2]+DX[2])),
                                np.array((X0[0]+DX[0], X0[1],       X0[2]+DX[2])),
                                np.array((X0[0]+DX[0], X0[1]+DX[1], X0[2]+DX[2])),
                                np.array((X0[0],       X0[1]+DX[1], X0[2]+DX[2]))))
        else:
            hopout.error('No corners or DX vector given for zone {}'.format(zone+1))

        nElems = GetIntArray(  'nElems'  , number=zone)
        # Store the requested element types
        if CountOption('ElemType') == 1 and zone > 0:
            elemTypes[zone] = elemTypes[0]
        else:
            elemTypes[zone] = GetIntFromStr('ElemType', number=zone)
        # ... but GMSH always builds hexahedral elements
        elemType = 108

        # Create all the corner points
        p = [None for _ in range(len(corners))]
        for index, corner in enumerate(corners):
            p[index] = gmsh.model.geo.addPoint(*cast(tuple[float, float, float], corner), tag=offsetp+index+1)

        # Define edge connectivity based on the Gmsh corner indexing
        edge_pairs = [
            (0, 1), (1, 2), (2, 3), (3, 0),  # Bottom face edges
            (4, 5), (5, 6), (6, 7), (7, 4),  # Top face edges
            (0, 4), (1, 5), (2, 6), (3, 7)   # Vertical edges
        ]

        # Connect the corner points
        e = [None for _ in range(12)]
        # First, the plane surface
        for i in range(2):
            for j in range(4):
                e[j + i*4] = gmsh.model.geo.addLine(p[j + i*4], p[(j+1) % 4 + i*4])
        # Then, the connection
        for j in range(4):
            e[j+8] = gmsh.model.geo.addLine(p[j], p[j+4])

        # Extract edge vectors from 'corners' for orientation checking
        edge_vectors = [corners[end] - corners[start] for start, end in edge_pairs]

        # Get dimensions of domain
        gmsh.model.geo.synchronize()
        box    = gmsh.model.get_bounding_box(-1, -1)
        lEdges = np.zeros([3])
        for i in range(3):
            lEdges[i] = np.abs(box[i+3]-box[i])

        # Get streching information of current zone
        stretchType = GetIntArray('StretchType', number=zone)

        # Check which stretching type is used and calculate the required factors
        if np.all(stretchType == 0) and (CountOption('l0') > 0 or CountOption('Factor') > 0):
            # No stretching however check if l0 or factor is gievn in parameter file
            # and assume that the user wants to use the factor stretching by default.
            # Calculate the stretching parameter for meshing the current zone
            stretchType[:] = 1
            # print warning which indicates that default value has been changed
            print(hopout.warn('Default StretchType changed for the current zone since '
                              'Factor or l0 is provided.'))

        # Progression factor stretching or double sided stretching
        stretchFac = np.ndarray([])
        if 1 in stretchType:
            stretchFac = CalcStretching(nZones, zone, nElems, lEdges)

        # Ratio based stretching
        DXmaxToDXmin = np.ndarray([])
        if 2 in stretchType or 3 in stretchType:
            DXmaxToDXmin = GetRealArray('DXmaxToDXmin', number=zone)

        maxStretch = np.array([1., 1., 1.], dtype=np.float128)
        for currDir in range(3):
            match stretchType[currDir]:
                case 1:
                    maxStretch[currDir] = abs(stretchFac[currDir]) ** nElems[currDir]
                case 2:
                    maxStretch[currDir] = -1./(DXmaxToDXmin[currDir] ** (1. / (nElems[currDir] - 1.)))
                case 3:
                    maxStretch[currDir] =  1./DXmaxToDXmin[currDir]

        # Stretching factor greater 2^26 ~ 10^8
        if np.max(maxStretch) > 2 << 26:  # pragma: no cover
            print(hopout.warn(f'Maximum stretching factor is {np.max(maxStretch)}!'))

        # We need to define the curves as transfinite curves
        # and set the correct spacing from the parameter file
        for index, line in enumerate(e):

            # We set the number of nodes, so Elems+1
            currDir  = edge_to_dir(index, elemType)

            # Set default values for equidistant elements
            progType = 'Progression'
            progFac  = 1.

            # Overwrite default values to consider streching in current zone
            match stretchType[currDir]:
                case 1:
                    progFac  = stretchFac[currDir]
                case 2:
                    progFac  = -1./(DXmaxToDXmin[currDir] ** (1. / (nElems[currDir] - 1.)))
                case 3:
                    progType = 'Bump'
                    progFac  =  1./DXmaxToDXmin[currDir]

            gmsh.model.geo.mesh.setTransfiniteCurve(line,
                                                    nElems[currDir]+1,
                                                    progType,
                                                    (np.sign(edge_vectors[index][currDir]) or 1.) * progFac)

        # Create the curve loop
        el = [None for _ in range(len(faces(elemType)))]
        for index, face in enumerate(faces(elemType)):
            el[index] = gmsh.model.geo.addCurveLoop([math.copysign(e[abs(s)], s) for s in face_to_edge(face, elemType)])

        # Create the surfaces
        s = [None for _ in range(len(faces(elemType)))]
        for index, _ in enumerate(s):
            s[index] = gmsh.model.geo.addPlaneSurface([el[index]], tag=offsets+index+1)

        # We need to define the surfaces as transfinite surface
        for index, face in enumerate(faces(elemType)):
            gmsh.model.geo.mesh.setTransfiniteSurface(offsets+index+1, face, [p[s] for s in face_to_corner(face, elemType)])
            gmsh.model.geo.mesh.setRecombine(2, 1)

        # Create the surface loop
        gmsh.model.geo.addSurfaceLoop([s for s in s], zone+1)

        gmsh.model.geo.synchronize()

        # Create the volume
        gmsh.model.geo.addVolume([zone+1], zone+1)

        # We need to define the volume as transfinite volume
        gmsh.model.geo.mesh.setTransfiniteVolume(zone+1)
        gmsh.model.geo.mesh.setRecombine(3, 1)

        # Calculate all offsets
        offsetp += len(corners)
        offsets += len(faces(elemType))

        # Read the BCs for the zone
        # > Need to wait with defining physical boundaries until all zones are created
        bcZones[zone] = [int(s) for s in GetIntArray('BCIndex')]

        # Assign the volume to a physical zone
        _ = gmsh.model.addPhysicalGroup(3, [zone+1], name='Zone{}'.format(zone+1))

    # At this point, we can create a "Physical Group" corresponding
    # to the boundaries. This requires a synchronize call!
    gmsh.model.geo.synchronize()

    hopout.sep()
    hopout.routine('Setting boundary conditions')
    hopout.sep()
    nBCs = CountOption('BoundaryName')
    mesh_vars.bcs = [BC() for _ in range(nBCs)]
    bcs = mesh_vars.bcs

    for iBC, bc in enumerate(bcs):
        # bcs[iBC].update(name = GetStr(     'BoundaryName', number=iBC),  # noqa: E251
        #                 bcid = iBC + 1,                                  # noqa: E251
        #                 type = GetIntArray('BoundaryType', number=iBC))  # noqa: E251
        bcs[iBC].name = GetStr(     'BoundaryName', number=iBC).lower()    # noqa: E251
        bcs[iBC].bcid = iBC + 1                                            # noqa: E251
        bcs[iBC].type = GetIntArray('BoundaryType', number=iBC)            # noqa: E251

    nVVs = CountOption('vv')
    if nVVs > 0:
        hopout.sep()
    mesh_vars.vvs = [dict() for _ in range(nVVs)]
    vvs = mesh_vars.vvs
    for iVV, _ in enumerate(vvs):
        vvs[iVV] = dict()
        vvs[iVV]['Dir'] = GetRealArray('vv', number=iVV)

    # Flatten the BC array, the surface numbering follows from the 2-D ordering
    bcIndex = [item for row in bcZones for item in row]

    bc = [None for _ in range(max(bcIndex))]
    for iBC in range(max(bcIndex)):
        # if mesh_vars.bcs[iBC-1] is None:
        # if 'Name' not in bcs[iBC]:
        if bcs[iBC] is None:
            continue

        # Format [dim of group, list, name)
        # > Here, we return ALL surfaces on the BC, irrespective of the zone
        surfID  = [s+1 for s in find_indices(bcIndex, iBC+1)]
        bc[iBC] = gmsh.model.addPhysicalGroup(2, surfID, name=cast(str, bcs[iBC].name))

        # For periodic sides, we need to impose the periodicity constraint
        if cast(np.ndarray, bcs[iBC].type)[0] == 1:
            # > Periodicity transform is provided as a 4x4 affine transformation matrix, given by row
            # > Rotation matrix [columns 0-2], translation vector [column 3], bottom row [0, 0, 0, 1]

            # Only define the positive translation
            if cast(np.ndarray, bcs[iBC].type)[3] > 0:
                pass
            elif cast(np.ndarray, bcs[iBC].type)[3] == 0:
                hopout.error('BC "{}" has no periodic vector given, exiting...'.format(iBC + 1), traceback=True)
            else:
                continue

            translation = [1., 0., 0., float(vvs[int(cast(np.ndarray, bcs[iBC].type)[3])-1]['Dir'][0]),
                           0., 1., 0., float(vvs[int(cast(np.ndarray, bcs[iBC].type)[3])-1]['Dir'][1]),
                           0., 0., 1., float(vvs[int(cast(np.ndarray, bcs[iBC].type)[3])-1]['Dir'][2]),
                           0., 0., 0., 1.]

            # Find the opposing side(s)
            # > copy, otherwise we modify bcs
            nbType     = cast(list, copy.copy(bcs[iBC].type))
            nbType[3] *= -1
            nbBCID     = find_index([s.type for s in bcs], nbType)
            # nbSurfID can hold multiple surfaces, depending on the number of zones
            # > find_indices returns all we need!
            nbSurfID   = [s+1 for s in find_indices(bcIndex, nbBCID+1)]

            # Connect positive to negative side
            try:
                gmsh.model.mesh.setPeriodic(2, nbSurfID, surfID, translation)
                hopout.routine('Generated periodicity constraint with vector {}'.format(
                    vvs[int(cast(np.ndarray, bcs[iBC].type)[3])-1]['Dir']))

            # If the number of sides do not match, we cannot impose periodicity
            # > Leave it out here and assume we can sort it out in ConnectMesh
            except Exception as e:
                print(hopout.warn(' No GMSH periodicity with vector {}'.format(
                    vvs[int(cast(np.ndarray, bcs[iBC].type)[3])-1]['Dir'])))
                continue

    if len(vvs) > 0:
        hopout.sep()

    # To generate connect the generated cells, we can simply set
    gmsh.option.setNumber('Mesh.RecombineAll'  , 1)
    gmsh.option.setNumber('Mesh.Recombine3DAll', 1)
    gmsh.option.setNumber('Geometry.AutoCoherence', 2)
    gmsh.model.mesh.recombine()
    # Force Gmsh to output all mesh elements
    gmsh.option.setNumber('Mesh.SaveAll', 1)

    gmsh.model.mesh.generate(3)

    # Set the element order
    # > This needs to be executed after generate_mesh, see
    # > https://github.com/nschloe/pygmsh/issues/515#issuecomment-1020106499
    gmsh.model.mesh.setOrder(mesh_vars.nGeo)
    gmsh.model.geo.synchronize()

    if debugvisu and IsDisplay():
        gmsh.fltk.run()
        # Re-set the order for newly created elements
        gmsh.model.mesh.setOrder(mesh_vars.nGeo)
        gmsh.model.geo.synchronize()

    # Consistency check if the mesh contains volume elements
    # > User might have modified the mesh inside the FLTK GUI
    # gmshElems = np.asarray((gmsh.option.getNumber('Mesh.NbTetrahedra'),
    #                         gmsh.option.getNumber('Mesh.NbPrisms'    ),
    #                         gmsh.option.getNumber('Mesh.NbPyramids'  ),
    #                         gmsh.option.getNumber('Mesh.NbHexahedra')), dtype=int)
    gmshTypes = gmsh.model.mesh.getElementTypes()
    gmshElems = np.asarray([(elemName, order) for type                          in gmshTypes                                     # noqa: E272
                                               for elemName, dim, order, _, _, _ in [gmsh.model.mesh.getElementProperties(type)]  # noqa: E272
                              if dim == 3])
    if not np.any(gmshElems):
        hopout.error('Generated mesh does not contain volume elements, exiting...')

    # Consistency check if the mesh elements have the correct order
    gmshIssue  = np.asarray([(elemName, order) for type                          in gmshTypes                                     # noqa: E272
                                               for elemName, dim, order, _, _, _ in [gmsh.model.mesh.getElementProperties(type)]  # noqa: E272
                              if dim == 3 and order != mesh_vars.nGeo])

    if gmshIssue.size > 0:
        for elem in gmshIssue:
            print(hopout.warn(f'Wrong Gmsh order {elem[1]} for element {elem[0].replace(" ", "")}'))
        elemOrders = set([int(elem[1]) for elem in gmshIssue])
        hopout.error(f'Gmsh element order(s) {elemOrders} does not match requested mesh order {set([mesh_vars.nGeo])}')

    # Convert Gmsh object to meshio object
    mesh = gmsh_to_meshio(gmsh)

    # Finally done with GMSH, finalize
    gmsh.finalize()

    # Run garbage collector to release memory
    gc.collect()

    # Store the element types
    mesh_vars.nZones    = nZones
    mesh_vars.elemTypes = elemTypes

    return mesh
